import os

import pytest

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

from angorapy.agent.ppo_agent import PPOAgent
from angorapy.common.policies import BetaPolicyDistribution, MultiCategoricalPolicyDistribution
from angorapy.common.postprocessors import RewardNormalizer, StateNormalizer
from angorapy import make_task

from angorapy.models import get_model_builder

try:
    from mpi4py import MPI

    is_root = MPI.COMM_WORLD.rank == 0
except:
    is_root = True

def _test_drill(env_name, model_builder=None, ):
    """Test drilling an agent."""
    wrappers = [StateNormalizer, RewardNormalizer]
    env = make_task(env_name, reward_config=None, transformers=wrappers)
    if model_builder is None:
        build_models = get_model_builder(model="simple", model_type="ffn", shared=False)
    else:
        build_models = model_builder
    agent = PPOAgent(build_models, env, workers=2, horizon=max(128, env.spec.max_episode_steps))
    agent.drill(n=2, epochs=2, batch_size=64)


def test_drill_continuous():
    """Test drilling of continuous agent (LunarLanderContinuous)."""

    try:
        _test_drill("LunarLanderContinuous-v2")
    except Exception:
        pytest.fail("Continuous drill raises error.")


def test_drill_discrete():
    """Test drilling of discrete agent (LunarLander)."""

    try:
        _test_drill("LunarLander-v2")
    except Exception:
        pytest.fail("Discrete drill raises error.")


def test_drill_manipulate_multicategorical():
    """Test drilling of discrete agent (LunarLander)."""

    try:
        wrappers = [StateNormalizer, RewardNormalizer]
        env = make_task("ManipulateBlockDiscreteAsynchronous-v0", reward_config=None, transformers=wrappers)
        build_models = get_model_builder(model="shadow", model_type="lstm", shared=False, blind=True)
        agent = PPOAgent(
            build_models,
            env,
            workers=2,
            horizon=128,
            distribution=MultiCategoricalPolicyDistribution(env)
        )

        agent.drill(n=2, epochs=2, batch_size=64)
    except Exception:
        pytest.fail("ManipulateBlockDiscreteAsynchronous drill raises error.")


def test_drill_manipulate_continuous():
    """Test drilling of discrete agent (LunarLander)."""

    try:
        wrappers = [StateNormalizer, RewardNormalizer]
        env = make_task("ManipulateBlockAsynchronous-v0", reward_config=None, transformers=wrappers)
        build_models = get_model_builder(model="shadow", model_type="lstm", shared=False)
        agent = PPOAgent(build_models, env, workers=2, horizon=128, distribution=BetaPolicyDistribution(env))
        agent.drill(n=2, epochs=2, batch_size=64)
    except Exception:
        pytest.fail("ManipulateBlockDiscreteAsynchronous drill raises error.")


def test_drill_reach():
    """Test drilling of discrete agent (LunarLander)."""

    try:
        wrappers = [StateNormalizer, RewardNormalizer]
        env = make_task("ReachAbsolute-v0", reward_config=None, transformers=wrappers)
        build_models = get_model_builder(model="shadow", model_type="lstm", shared=False)
        agent = PPOAgent(build_models, env, workers=2, horizon=128, distribution=BetaPolicyDistribution(env))
        agent.drill(n=2, epochs=2, batch_size=64)
    except Exception:
        pytest.fail("ReachAbsolute drill raises error.")


def test_drill_freereach():
    """Test drilling of free reach agent."""

    try:
        wrappers = [StateNormalizer, RewardNormalizer]
        env = make_task("FreeReachAbsolute-v0", reward_config=None, transformers=wrappers)
        build_models = get_model_builder(model="shadow", model_type="lstm", shared=False)
        agent = PPOAgent(build_models, env, workers=2, horizon=128, distribution=BetaPolicyDistribution(env))
        agent.drill(n=2, epochs=2, batch_size=64)
    except Exception:
        pytest.fail("FreeReachAbsolute drill raises error.")


def test_classic_control():
    for env_name in ["CartPole-v1", "Acrobot-v1", "Pendulum-v1", "MountainCar-v0"]:
        try:
            _test_drill(env_name)
        except Exception:
            pytest.fail("Continuous drill raises error.")


def test_robotic_control():
    for env_name in ["Ant-v4", "Humanoid-v4"]:
        try:
            _test_drill(env_name)
        except Exception:
            pytest.fail("Continuous drill raises error.")



import re

from matplotlib import pyplot as plt

with open("results/predictability_.txt", "r") as f:
    results = [re.match("(.*) from (.*) has an R2 of (.*)", line) for line in f.readlines()]

results = [r.groups() for r in results if r is not None]
result_dict = {}

for r_target, r_activity, r2 in results:
    if r_target not in result_dict:
        result_dict[r_target] = {}

    result_dict[r_target][r_activity] = float(r2[:-1])


for target, activity_r2s in result_dict.items():
    source_names = activity_r2s.keys()
    source_r2s = [activity_r2s[n] for n in source_names]
    plt.bar(source_names, source_r2s)
    plt.title(target)
    plt.xticks(rotation="vertical")
    plt.gcf().subplots_adjust(bottom=0.3)
    plt.show()



import json
import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from dexterity.design.common import QUALITATIVE_COLOR_PALETTE

AGENT_ID =  1581809147

with open(f"../../monitor/static/experiments/{AGENT_ID}/progress.json", "r") as f:
    data = json.load(f)

vloss = data["vloss"]
ploss = data["ploss"]
entropy = data["entropies"]
rewards = data["rewards"]["mean"]

fig: Figure
fig, axs = plt.subplots(1, 3)
fig.set_size_inches(16, 4)

# axs[0].set_title("Policy Gradient Loss")
# axs[1].set_title("Value Function Loss")
# axs[2].set_title("Entropy Bonus")

# plot rewards
for ax in axs:
    ax.set_zorder(10)
    ax.patch.set_visible(False)
    twin_ax = ax.twinx()
    pr = twin_ax.plot(rewards, color="lightgrey", zorder=-1)
    twin_ax.set_ylabel("Mean Cumulative Reward")

p1 = axs[0].plot(ploss, color=QUALITATIVE_COLOR_PALETTE[0], zorder=10)
p2 = axs[1].plot(vloss, color=QUALITATIVE_COLOR_PALETTE[1], zorder=10)
p3 = axs[2].plot(entropy, color=QUALITATIVE_COLOR_PALETTE[2], zorder=10)

axs[0].set_ylabel("Cycle Mean Loss")
axs[1].set_ylabel("Cycle Mean Loss")
axs[2].set_ylabel("Cycle Mean Bonus")
fig.legend([p1, p2, p3, pr],
           labels=["Policy Gradient Loss", "Value Gradient Loss", "Entropy Bonus", "Reward"],
           loc="lower center", bbox_to_anchor=(0.5, 0),
           ncol=4)

plt.subplots_adjust(wspace=.5, bottom=0.2)
plt.savefig("../../docs/figures/losses.pdf", format="pdf", bbox_inches='tight')
plt.show()



import itertools
import json
import os
import re
from json import JSONDecodeError
from typing import Iterable

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

from angorapy.common.const import PATH_TO_EXPERIMENTS, BASE_SAVE_PATH, QUALITATIVE_COLOR_PALETTE

matplotlib.use('TkAgg')

group_names = {
    "gaussian": [
        "benchmark-performance-ant",
        "benchmark-performance-walker2d",
        "benchmark-performance-swimmer",
        "benchmark-gaussian-reach",
        "benchmark-performance-reacher",
        "benchmark-performance-hopper",
        "benchmark-performance-halfcheetah",
        "benchmark-gaussian-freereach"
    ], "beta": [
        "benchmark-beta-ant",
        "benchmark-beta-walker2d",
        "benchmark-beta-swimmer",
        "benchmark-beta-reach",
        "benchmark-beta-reacher",
        "benchmark-beta-hopper",
        "benchmark-beta-halfcheetah",
        "benchmark-beta-freereach"
    ]
}

titles = [n.split("-")[-1].capitalize() for n in group_names[list(group_names.keys())[0]]]

# group_names = {"any": ["benchmark-performance-pdl", "benchmark-performance-llc", "benchmark-performance-cp",
#                "benchmark-performance-ab"]}
# titles = ["Pendulum", "LunarLanderContinuous", "CartPole", "Acrobot"]

exp_dir = "../../../" + PATH_TO_EXPERIMENTS
experiment_paths = [os.path.join(exp_dir, p) for p in os.listdir(exp_dir)]

environments = {category: {} for category in group_names.keys()}
reward_thresholds = {category: {} for category in group_names.keys()}
experiments_by_groups = {category: {} for category in group_names.keys()}
envs_available = set()

for exp_path in experiment_paths:

    eid_m = re.match("[0-9]+", str(exp_path.split("/")[-1]))
    if eid_m:
        eid = eid_m.group(0)
        model_path = os.path.join(BASE_SAVE_PATH, eid)

        if os.path.isfile(os.path.join(exp_path, "progress.json")):
            with open(os.path.join(exp_path, "progress.json"), "r") as f:
                progress = json.load(f)

            with open(os.path.join(exp_path, "meta.json"), "r") as f:
                try:
                    meta = json.load(f)
                except JSONDecodeError as jserr:
                    continue

            exp_group = meta.get("experiment_group", "n/a")

            if exp_group not in itertools.chain(*group_names.values()):
                continue

            reward_threshold = None if meta["environment"]["reward_threshold"] == "None" else float(
                meta["environment"]["reward_threshold"])

            for category in group_names.keys():
                if exp_group in group_names[category] and exp_group not in experiments_by_groups[category].keys():
                    experiments_by_groups[category][exp_group] = {}
                    reward_thresholds[category][exp_group] = reward_threshold
                    environments[category][exp_group] = meta["environment"]["name"]

            envs_available.add(meta["environment"]["name"])

            for category in group_names.keys():
                if exp_group in group_names[category]:
                    experiments_by_groups[category][exp_group].update({
                        eid: progress
                    })

n_rows, n_cols = 2, 4
fig, axs = plt.subplots(n_rows, n_cols)
fig.set_size_inches(16, 4 * n_rows)

if not isinstance(axs[0], Iterable):
    axs = [axs]


for i_cat, category in enumerate(group_names.keys()):
    for i, name in enumerate(group_names[category]):
        data = experiments_by_groups[category][name]
        reward_trajectories = list(map(lambda x: x["rewards"]["mean"], data.values()))
        max_length = max([len(x) for x in reward_trajectories])
        padded_reward_trajectories = list(map(lambda x: np.pad(x, (0, max_length - len(x)),
                                                               mode="constant",
                                                               constant_values=np.nan), reward_trajectories))
        mean_reward = np.ma.mean(np.ma.array(padded_reward_trajectories, mask=np.isnan(padded_reward_trajectories)),
                                 axis=0)
        std_reward = np.ma.std(np.ma.array(padded_reward_trajectories, mask=np.isnan(padded_reward_trajectories)),
                               axis=0)

        ax: Axes = axs[i // n_cols][i % n_cols]

        # if reward_thresholds[category][name] is not None:
        #     ax.axhline(reward_thresholds[category][name], color=QUALITATIVE_COLOR_PALETTE[2], ls="--")
        ax.plot(mean_reward, label=category, color=QUALITATIVE_COLOR_PALETTE[i_cat])
        ax.fill_between(range(mean_reward.shape[0]), mean_reward - std_reward, mean_reward + std_reward, alpha=.2)

        ax.set_xlim(0, mean_reward.shape[0] - 1)
        ax.set_ylim(min(np.min(mean_reward - std_reward), ax.get_ylim()[0]),
                    max(np.max(mean_reward + std_reward) * 1.1, ax.get_ylim()[1]))
        ax.set_xlabel("Cycle")
        ax.set_title(titles[i])

        if titles[i] in ["Reach", "Freereach"]:
            ax.set_title(titles[i], fontstyle="italic")

        if i % n_cols == 0:
            ax.set_ylabel("Episode Return")

        if len(group_names.keys()) > 1:
            ax.legend(loc="lower right")

plt.subplots_adjust(top=0.8, bottom=0.2, hspace=0.35, wspace=0.2)

plt.savefig(f"../../../docs/figures/benchmarks/{'_'.join(titles)}_benchmark_comparison.pdf", format="pdf",
            bbox_inches='tight')
plt.show()



import tensorflow as tf
import tensorflow_graphics as tfg


def multi_point_euclidean_distance(a: tf.Tensor, b: tf.Tensor):
    """Calculate mean euclidean distance between flattened points in 3D euclidean space.

    Args:
        a (object):     Tensor of shape (BATCH, FLATPOINTS); missing batch dimension will be inserted
        b (object):     Tensor of shape (BATCH, FLATPOINTS); missing batch dimension will be inserted
    """
    assert a.shape.as_list() == b.shape.as_list(), f"Cannot measure euclidean distance between tensors of unequal shape ({a.shape} != {b.shape})."
    assert a.shape[-1] % 3 == 0, "Flattened point dimension must be multiple of 3 to get 3D distance."

    if len(a.shape) < 2:
        a = tf.expand_dims(a, 0)
        b = tf.expand_dims(b, 0)

    batch_dim = tf.shape(a)[0]

    return tf.reduce_mean(
        tf.sqrt(
            tf.reduce_sum(
                tf.square(
                    tf.subtract(
                        tf.reshape(a, (batch_dim, -1, 3)),
                        tf.reshape(b, (batch_dim, -1, 3))
                    )
                ), axis=-1
            )
        )
    )


class EuclideanDistanceLoss(tf.keras.losses.Loss):

    def call(self, y_true, y_pred):
        y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)

        return multi_point_euclidean_distance(y_pred, y_true)


class PoseEstimationLoss(tf.keras.losses.Loss):

    beta = 0.001

    def call(self, y_true, y_pred):
        y_pred = tf.convert_to_tensor(y_pred)
        y_true = tf.cast(y_true, y_pred.dtype)

        position_pred = y_pred[:, :3]
        position_true = y_true[:, :3]

        rotation_pred = y_pred[:, 3:]
        rotation_true = y_true[:, 3:]

        # find the euclidean distance between the two positions in the pose
        position_loss = tf.reduce_mean(tf.norm(position_pred - position_true, ord="euclidean", axis=-1))

        # find the shortest geodesic rotation between the two quaternions in the pose
        batch_dot_product = tf.reduce_sum(rotation_pred * rotation_true, axis=-1)
        rotational_loss = tf.reduce_mean(1.0 - tf.pow(batch_dot_product, 2))

        return position_loss + rotational_loss



"""Generation of auxiliary datasets."""
import os

import numpy as np
import tensorflow as tf
from tqdm import tqdm

from angorapy import make_task
from angorapy.agent import PPOAgent
from angorapy.common.const import VISION_WH


def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()  # BytesList won't unpack a string from an EagerTensor.
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def serialize_example(image, pose):
    feature = {
        'sim_state': _bytes_feature(image),
        'pose': _bytes_feature(pose),
    }

    example_proto = tf.train.Example(features=tf.train.Features(feature=feature))
    return example_proto.SerializeToString()


def load_dataset(path):
    feature_description = {
        "sim_state": tf.io.FixedLenFeature([], tf.string),
        "pose": tf.io.FixedLenFeature([], tf.string),
    }

    def _parse_function(example_proto):
        # Parse the input `tf.Example` proto using the dictionary above.
        parsed = tf.io.parse_single_example(example_proto, feature_description)
        parsed["sim_state"] = tf.io.parse_tensor(parsed["sim_state"], out_type=tf.float32)
        parsed["pose"] = tf.io.parse_tensor(parsed["pose"], out_type=tf.float32)

        return tf.reshape(parsed["sim_state"], (VISION_WH, VISION_WH, -1)), tf.reshape(parsed["pose"], (7,))

    return tf.data.TFRecordDataset(path).map(_parse_function)


def load_unrendered_dataset(path):
    feature_description = {
        "sim_state": tf.io.FixedLenFeature([], tf.string),
        "pose": tf.io.FixedLenFeature([], tf.string),
    }

    def _parse_function(example_proto):
        # Parse the input `tf.Example` proto using the dictionary above.
        parsed = tf.io.parse_single_example(example_proto, feature_description)
        parsed["sim_state"] = tf.io.parse_tensor(parsed["sim_state"], out_type=tf.float32)
        parsed["pose"] = tf.io.parse_tensor(parsed["pose"], out_type=tf.float32)

        return tf.reshape(parsed["sim_state"], (-1,)), tf.reshape(parsed["pose"], (7,))

    return tf.data.TFRecordDataset(path).map(_parse_function)


def gen_cube_quats_prediction_data(n: int, save_path: str):
    """Generate dataset of hand images with cubes as data points and the pos and rotation of the cube as targets."""
    agent = PPOAgent.from_agent_state(1692003322304510, "best", path_modifier="")
    agent.policy, agent.value, agent.joint = agent.build_models(agent.joint.get_weights(), batch_size=1,
                                                                sequence_length=1)

    hand_env = agent.env

    os.makedirs("storage/data/pretraining", exist_ok=True)
    with tf.io.TFRecordWriter(save_path) as writer:
        state, _ = hand_env.reset()
        for i in tqdm(range(n), desc="Sampling Data"):
            action = agent.act(state)
            sample, r, terminated, truncated, info = hand_env.step(action)
            done = terminated or truncated

            quaternion = tf.cast(info["achieved_goal"], dtype=tf.float32)
            sim_state = hand_env.get_state()
            sim_state = tf.cast(np.concatenate([sim_state["qpos"], sim_state["qvel"]]), dtype=tf.float32)

            example = serialize_example(tf.io.serialize_tensor(sim_state), tf.io.serialize_tensor(quaternion))
            writer.write(example)

            if done == 0:
                sample, _ = hand_env.reset()

            state = sample


def gen_cube_quats_prediction_data_rendered(n: int, save_path: str, binocular=False, n_cameras: int = 1):
    """Generate dataset of hand images with cubes as data points and the pos and rotation of the cube as targets."""
    agent = PPOAgent.from_agent_state(1670596186987840, "best", path_modifier="../")
    agent.policy, agent.value, agent.joint = agent.build_models(agent.joint.get_weights(), batch_size=1,
                                                                sequence_length=1)

    hand_env = make_env("HumanoidManipulateBlockDiscreteAsynchronous-v0",
                        # render_mode="rgb_array",
                        transformers=agent.env.transformers)

    # env_unwrapped: AnthropomorphicEnv = hand_env.unwrapped
    # model = env_unwrapped.model
    # data = env_unwrapped.data
    # renderer = mujoco.Renderer(model, height=VISION_WH, width=VISION_WH)
    # cameras = [env_unwrapped._get_viewer("rgb_array").cam]
    # for i in range(1, n_cameras):
    #     cameras.append(mujoco.MjvCamera())
    #     cameras[-1].type = mujoco.mjtCamera.mjCAMERA_FREE
    #     cameras[-1].fixedcamid = -1
    #     for j in range(3):
    #         cameras[-1].lookat[j] = np.median(data.geom_xpos[:, j])
    #     cameras[-1].distance = model.stat.extent
    #
    #     cameras[-1].distance = 0.3  # zoom in
    #     cameras[-1].azimuth = -90.0 + [20, -20][i - 1]  # wrist to the bottom
    #     cameras[-1].elevation = -90.0 + [20, -20][i - 1]  # top down view
    #     cameras[-1].lookat[1] += 0.03  # slightly move forward
    #
    # def random_augmentation():
    #     # Choose a random number between 0 and 1
    #     p = np.random.random()
    #
    #     # Get the quaternion for the current orientation of the body
    #     object = data.jnt("object:joint")
    #     quat = data.jnt("object:joint").qpos[3:]
    #     original_qpos = object.qpos.copy()
    #
    #     if p < 0.0:  # Leave the object pose as is
    #         pass
    #     elif p < 1:  # Rotate the object by 90 degrees around its main axes
    #         # define the rotation
    #         axis = np.random.choice(['x', 'y', 'z'])
    #         sign = np.random.choice([-1, 1])
    #
    #         # rotate the body by 90 degrees around the chosen axis
    #         if axis == 'x':
    #             rotation_quat = tfg.quaternion.from_axis_angle(np.array([[1., 0., 0.]]), np.array([[sign * np.pi / 2]]))
    #         elif axis == 'y':
    #             rotation_quat = tfg.quaternion.from_axis_angle(np.array([[0., 1., 0.]]), np.array([[sign * np.pi / 2]]))
    #         else:
    #             rotation_quat = tfg.quaternion.from_axis_angle(np.array([[0., 0., 1.]]), np.array([[sign * np.pi / 2]]))
    #
    #         # Apply the rotation to the body quaternion
    #         new_quat = tfg.quaternion.multiply(quat, rotation_quat)
    #         object.qpos[3:] = new_quat
    #     else:  # "Jitter" the object by adding Gaussian noise to position and orientation
    #         # Add Gaussian noise to the body position
    #         noise_pos = np.random.normal(loc=0, scale=0.01, size=3)
    #         object.qpos[:3] += noise_pos
    #
    #         # Add Gaussian noise to the body orientation
    #         noise_quat = np.random.normal(loc=0, scale=0.01, size=4)
    #         noise_quat /= np.linalg.norm(noise_quat)
    #         new_quat = tfg.quaternion.multiply(quat, noise_quat)
    #         data.qpos[3:] = new_quat
    #
    #     mujoco.mj_forward(model, data)
    #
    #     images = []
    #     for cam in cameras:
    #         renderer.update_scene(data, camera=cam)
    #         images.append(renderer.render().copy())
    #
    #     qpos = object.qpos.copy()
    #     object.qpos = original_qpos
    #
    #     # plt.imshow(np.concatenate(images, axis=-2))
    #     # plt.show()
    #
    #     return tf.cast(tf.concat(images, axis=-1) / 255, dtype=tf.float32), tf.cast(qpos, tf.float32)

    os.makedirs("storage/data/pretraining", exist_ok=True)
    # with tf.io.TFRecordWriter(save_path) as writer:
    # with None as writer:
    state, _ = hand_env.reset()
    for i in tqdm(range(n), desc="Sampling Data"):
        # state["vision"] = hand_env.data.jnt('object:joint').qpos.copy()

        action = agent.act(state, joint=agent.joint)
        sample, r, terminated, truncated, info = hand_env.step(action)
        done = terminated or truncated
        # image, quaternion = random_augmentation()
        image, quaternion = sample["vision"], info["achieved_goal"]
        # X[i] = image
        # Y[i] = quaternion

        # example = serialize_example(tf.io.serialize_tensor(image), tf.io.serialize_tensor(quaternion))
        # writer.write(example)

        if done or i % 64 == 0:
            sample, _ = hand_env.reset()

        state = sample

    return load_dataset(save_path)


if __name__ == '__main__':
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    data = gen_cube_quats_prediction_data(1000, "../storage/data/pretraining/pose_data.tfrecord")
    print(data)



from typing import List

import numpy
import numpy as np
import tensorflow as tf


class Sensation(dict):
    """Wrapper for state representations. Designed for sensory readings, but non-sensory data can be
    encapsulated too. The convention is to refer to any non-sensory data as proprioception, following the intuition
    that it represents inner states."""

    sense_names = ["vision", "touch", "proprioception", "goal", "asymmetric"]

    def __init__(self, vision: np.ndarray = None, proprioception: np.ndarray = None, touch: np.ndarray = None,
                 goal: np.ndarray = None, asymmetric: np.ndarray = None):
        super().__init__()

        assert not all(s is None for s in (vision, touch, proprioception, goal)), \
            "Sensation requires data for at least on sense (or 'goal')."

        self.vision = vision.astype(np.float32) if vision is not None else None
        self.touch = touch.astype(np.float32) if touch is not None else None
        self.proprioception = proprioception.astype(np.float32) if proprioception is not None else None
        self.goal = goal.astype(np.float32) if goal is not None else None
        self.asymmetric = asymmetric.astype(np.float32) if asymmetric is not None else None

    def __repr__(self):
        return str(self.__dict__)

    # PROPERTIES

    @property
    def shape(self):
        return (99,)

    @property
    def v(self):
        """Shortcut for vision."""
        return self.vision

    @property
    def s(self):
        """Shortcut for touch."""
        return self.touch

    @property
    def t(self):
        """Shortcut for touch."""
        return self.touch

    @property
    def p(self):
        """Shortcut for proprioception."""
        return self.proprioception

    @property
    def g(self):
        """Shortcut for goal."""
        return self.goal

    # MAGIC

    def __iter__(self):
        return iter((self.vision, self.touch, self.proprioception, self.goal, self.asymmetric))

    def __len__(self):
        return len(self.dict())

    def __list__(self):
        return list(self.dict())

    def __contains__(self, item):
        return item in self.dict()

    def __getitem__(self, item):
        return self.dict()[item]

    def __setitem__(self, key, value):
        if key in self.__dict__.keys():
            self.__dict__[key] = value
        else:
            raise ValueError(f"{key} is not a sense")

    def items(self):
        return self.dict().items()

    # MAGIC MATH
    def __add__(self, other):
        new_sensation = Sensation(
            vision=self.vision + other.vision if self.vision is not None else None,
            proprioception=self.proprioception + other.proprioception if self.proprioception is not None else None,
            touch=self.touch + other.touch if self.touch is not None else None,
            goal=self.goal + other.goal if self.goal is not None else None,
            asymmetric=self.asymmetric + other.asymmetric if self.asymmetric is not None else None,
        )

        return new_sensation

    def __sub__(self, other):
        new_sensation = Sensation(
            vision=self.vision - other.vision if self.vision is not None else None,
            proprioception=self.proprioception - other.proprioception if self.proprioception is not None else None,
            touch=self.touch - other.touch if self.touch is not None else None,
            goal=self.goal - other.goal if self.goal is not None else None,
            asymmetric=self.asymmetric - other.asymmetric if self.asymmetric is not None else None,
        )

        return new_sensation

    def __pow__(self, other):
        new_sensation = Sensation(
            vision=self.vision ** other if self.vision is not None else None,
            proprioception=self.proprioception ** other if self.proprioception is not None else None,
            touch=self.touch ** other if self.touch is not None else None,
            goal=self.goal ** other if self.goal is not None else None,
            asymmetric=self.asymmetric ** other if self.asymmetric is not None else None,
        )

        return new_sensation

    def __truediv__(self, other):
        if isinstance(other, Sensation):
            new_sensation = Sensation(
                vision=self.vision / other.vision if self.vision is not None else None,
                proprioception=self.proprioception / other.proprioception if self.proprioception is not None else None,
                touch=self.touch / other.touch if self.touch is not None else None,
                goal=self.goal / other.goal if self.goal is not None else None,
                asymmetric=self.asymmetric / other.asymmetric if self.asymmetric is not None else None,
            )
        elif isinstance(other, (int, float)) or np.isscalar(other):
            new_sensation = Sensation(
                vision=self.vision / other if self.vision is not None else None,
                proprioception=self.proprioception / other if self.proprioception is not None else None,
                touch=self.touch / other if self.touch is not None else None,
                goal=self.goal / other if self.goal is not None else None,
                asymmetric=self.asymmetric / other if self.asymmetric is not None else None,
            )
        else:
            raise TypeError(f"Unsupported division of Sensation by {type(other)}")

        return new_sensation

    def keys(self):
        return self.dict().keys()

    def sqrt(self):
        new_sensation = Sensation(
            vision=np.sqrt(self.vision) if self.vision is not None else None,
            proprioception=np.sqrt(self.proprioception) if self.proprioception is not None else None,
            touch=np.sqrt(self.touch) if self.touch is not None else None,
            goal=np.sqrt(self.goal) if self.goal is not None else None,
            asymmetric=np.sqrt(self.asymmetric) if self.asymmetric is not None else None,
        )

        return new_sensation

    # OTHER

    def inject_leading_dims(self, time=False):
        """Expand state (inplace) to have a batch and/or time dimension."""
        sense: numpy.ndarray

        for sense, value in self.dict().items():
            if value is None:
                continue

            self[sense] = np.expand_dims(value, axis=(0 if not time else (0, 1)))

    def with_leading_dims(self, time=False):
        """Return a new state with batch and/or time dimension but keep this Sensation as is."""
        sense: numpy.ndarray

        new_sensation = Sensation(**(self.dict()))
        for sense, value in new_sensation.dict().items():
            if value is None:
                continue

            new_sensation[sense] = np.expand_dims(value, axis=(0 if not time else (0, 1)))

        return new_sensation

    def dict(self):
        """Return a dict of the senses and their values."""
        return {k: self.__dict__[k] for k in Sensation.sense_names if self.__dict__[k] is not None}

    def dict_as_tf(self):
        """Return dict as tf tensor."""
        return {k: tf.convert_to_tensor(self.__dict__[k]) for k in Sensation.sense_names if self.__dict__[k] is not None}


def stack_sensations(sensations: List[Sensation], add_batch_dim=False):
    """Stack Sensation objects over a prepended temporal domain."""
    if not add_batch_dim:
        return Sensation(**{
            sense: np.stack([s[sense] for s in sensations], axis=0) for sense in sensations[0].dict().keys()
        })
    else:
        return Sensation(**{
            sense: np.expand_dims(np.stack([s[sense] for s in sensations], axis=0), 0)
            for sense in sensations[0].dict().keys()
        })


if __name__ == '__main__':
    ss = stack_sensations([
        Sensation(goal=np.array([1, 2, 3, 4])),
        Sensation(goal=np.array([0, 1, 2, 3])),
        Sensation(goal=np.array([1, 1, 1, 1])),
    ])

    print(ss)



"""Implementation from https://github.com/RobertTLange/gym-hanoi/blob/master/gym_hanoi/envs/hanoi_env.py"""

import itertools
import random
from typing import Union

import gymnasium as gym
import numpy as np
from gymnasium import spaces

from angorapy.common.senses import Sensation
from angorapy.tasks.utils import convert_observation_to_space


class HanoiEnv(gym.Env):
    metadata = {'render.modes': ['human']}

    def __init__(self, render_mode=None):
        self.num_disks = 3
        self.env_noise = 0

        self.current_state: Union[None, np.ndarray] = None
        self.goal_state: Union[None, np.ndarray] = np.array(self.num_disks * (2,))

        self.done = None
        self.ACTION_LOOKUP = {0: "(0,1) - top disk of pole 0 to top of pole 1 ",
                              1: "(0,2) - top disk of pole 0 to top of pole 2 ",
                              2: "(1,0) - top disk of pole 1 to top of pole 0",
                              3: "(1,2) - top disk of pole 1 to top of pole 2",
                              4: "(2,0) - top disk of pole 2 to top of pole 0",
                              5: "(2,1) - top disk of pole 2 to top of pole 1"}

        self.action_space = spaces.Discrete(6)
        self.observation_space = convert_observation_to_space(self.reset()[0])

    def step(self, action):
        """
        * Inputs:
            - action: integer from 0 to 5 (see ACTION_LOOKUP)
        * Outputs:
            - current_state: state after transition
            - reward: reward from transition
            - done: episode state
            - info: dict of booleans (noisy?/invalid action?)
        0. Check if transition is noisy or not
        1. Transform action (0 to 5 integer) to tuple move - see Lookup
        2. Check if move is allowed
        3. If it is change corresponding entry | If not return same state
        4. Check if episode completed and return
        """
        if self.done:
            raise RuntimeError("Episode has finished. Call env.reset() to start a new episode.")

        info = {"transition_failure": False,
                "invalid_action": False}

        if self.env_noise > 0:
            r_num = random.random()
            if r_num <= self.env_noise:
                action = random.randint(0, self.action_space.n - 1)
                info["transition_failure"] = True

        move = action_to_move[action]

        if self.move_allowed(move):
            disk_to_move = min(self.disks_on_peg(move[0]))
            moved_state = list(self.current_state)
            moved_state[disk_to_move] = move[1]
            self.current_state = np.array(tuple(moved_state))
        else:
            info["invalid_action"] = True

        if np.all(self.current_state == self.goal_state):
            reward = 100
            self.done = True
        elif info["invalid_action"] == True:
            reward = -1
        else:
            reward = -0.01

        return self._get_obs(), reward, self.done, self.done, info

    def _get_obs(self):
        return {
            "observation": Sensation(
                proprioception=np.array(self.current_state),
            ),
        }

    def disks_on_peg(self, peg):
        """
        * Inputs:
            - peg: pole to check how many/which disks are in it
        * Outputs:
            - list of disk numbers that are allocated on pole
        """
        return [disk for disk in range(self.num_disks) if self.current_state[disk] == peg]

    def move_allowed(self, move):
        """
        * Inputs:
            - move: tuple of state transition (see ACTION_LOOKUP)
        * Outputs:
            - boolean indicating whether action is allowed from state!
        move[0] - peg from which we want to move disc
        move[1] - peg we want to move disc to
        Allowed if:
            * discs_to is empty (no disc of peg) set to true
            * Smallest disc on target pole larger than smallest on prev
        """
        disks_from = self.disks_on_peg(move[0])
        disks_to = self.disks_on_peg(move[1])

        if disks_from:
            return (min(disks_to) > min(disks_from)) if disks_to else True
        else:
            return False

    def reset(self, **kwargs):
        self.current_state = np.array(self.num_disks * (0,))
        self.done = False

        return self._get_obs(), {}

    def render(self, mode='human', close=False):
        return

    def get_movability_map(self, fill=False):
        # Initialize movability map
        mov_map = np.zeros(self.num_disks * (3,) + (6,))

        if fill:
            # Get list of all states as tuples
            id_list = self.num_disks * [0] + self.num_disks * [1] + self.num_disks * [2]
            states = list(itertools.permutations(id_list, self.num_disks))

            for state in states:
                for action in range(6):
                    move = action_to_move[action]
                    disks_from = []
                    disks_to = []
                    for d in range(self.num_disks):
                        if state[d] == move[0]:
                            disks_from.append(d)
                        elif state[d] == move[1]:
                            disks_to.append(d)

                    if disks_from:
                        valid = (min(disks_to) > min(disks_from)) if disks_to else True
                    else:
                        valid = False

                    if not valid: mov_map[state][action] = -np.inf

                    move_from = [m[0] for m in action_to_move]
                    move_to = [m[1] for m in action_to_move]

        # # Try to get rid of action loop - vectorize...
        # for state in states:
        #     s = np.array(state)
        #     disks_from = []
        #     disks_to = []
        #
        #     for d in range(self.num_disks):
        #         a_from = [a for a, v in enumerate(move_from) if v == s[d]]
        #         a_to = [a for a, v in enumerate(move_to) if v == s[d]]
        #
        #         if disks_from:
        #             valid = (min(disks_to) > min(disks_from)) if disks_to else True
        #         else:
        #             valid = False
        #
        #         if not valid:
        #             mov_map[state][action] = -np.inf
        return mov_map


action_to_move = [(0, 1), (0, 2), (1, 0),
                  (1, 2), (2, 0), (2, 1)]



REACH_BASE = dict(
    FORCE_MULTIPLIER=.0,  # scales punishment for force application
    SUCCESS_DISTANCE=.02,  # fingertip distance at which reach is a success
    SUCCESS_BONUS=.1,

    # ZONING and WEIGHTHING for auxiliary fingers; default values guarantee a max reward of ~0.009 per step for this
    AUXILIARY_ZONE_RADIUS=0.06015,  # this is the initial distance to the forefinger
    AUXILIARY_PENALTY_MULTIPLIER=0.01
)
MANIPULATE_BASE = dict(
    SUCCESS_BONUS=5,
    DROPPING_PENALTY=20,
    FORCE_MULTIPLIER=.0,
    TENDON_STRESS_MULTIPLIER=.0,
)
reach = dict(
    default=dict(REACH_BASE),
    force_punished=dict(REACH_BASE, FORCE_MULTIPLIER=0.05),
)
sequential_reach = dict(
    default=dict(REACH_BASE,
                 SUCCESS_DISTANCE=.5),
    great_success=dict(REACH_BASE,
                       SUCCESS_MULTIPLIER=5)
)
free_reach = dict(
    default=REACH_BASE,
    force_punished=dict(REACH_BASE, FORCE_MULTIPLIER=0.05),
    force_punished_light=dict(REACH_BASE, FORCE_MULTIPLIER=0.001),
)
free_reach_positive_reinforcement = dict(
    default=REACH_BASE,
    force_punished=dict(REACH_BASE,
                        FORCE_MULTIPLIER=0.05),
)
sequential_free_reach = dict(
    default=REACH_BASE,
    great_success=dict(REACH_BASE,
                       SUCCESS_MULTIPLIER=20),
    gentle_success=dict(REACH_BASE,
                        SUCCESS_DISTANCE=0.025)
)
manipulate = dict(
    default=MANIPULATE_BASE,
    penalized_force=dict(MANIPULATE_BASE,
                         FORCE_MULTIPLIER=0.001, ),
    penalized_anatomy=dict(MANIPULATE_BASE,
                           TENDON_STRESS_MULTIPLIER=0.01,
                           FORCE_MULTIPLIER=0.01, )
)



import tensorflow as tf


def sindy_ae_reconstruction_loss(inputs, reconstruction):
    return tf.keras.losses.mean_squared_error(inputs, reconstruction)


def sindy_ae_loss_in_x(dx, z_coordinates, ):
    pass


def sindy_ae_loss_in_z():
    pass


def sindy_regularization():
    pass



import json
import re

import numpy as np
from matplotlib import pyplot as plt
import matplotlib
font = {'family' : 'Times',
        'weight' : 'normal',
        'size'   : 12}

partial_results = []
for filename in ["./storage/predictability_repeated_august19d.json", "./storage/predictability_repeated_august19e.json"]:
    with open(filename, "r") as f:
        partial_results.append(json.load(f))

results = {}
for key in partial_results[0].keys():
    for target in partial_results[0][key].keys():
        for part in partial_results:
            if key not in part or target not in part[key]:
                continue

            if key not in results:
                results[key] = {}
            if target not in results[key]:
                results[key][target] = []

            results[key][target] = results[key][target] + part[key][target]

included_regions = ["m1_internal", "pmc_recurrent_layer", "goal"]
names = ["M1", "PMC", "Goal"]
included_targets = ["translation", "translation_to_10", "translation_to_50"]
# included_regions = ["M1_activation", "pmc_recurrent_layer"]
# included_targets = ["reward", "goal"]
# included_regions = ["M1_activation", "pmc_recurrent_layer"]
# included_targets = ['rotation_matrix', 'rotation_matrix_last_10']


x = np.arange(len(included_targets))
width = 0.8 / len(included_regions)

rects = []
fig, ax = plt.subplots()

for i, source in enumerate(included_regions, start=-1):
       ax.bar(x + i * width,
              [np.mean(results[source][t]) for t in included_targets],
              width,
              yerr=[np.std(results[source][t]) for t in included_targets],
              label=names[i + 1],
              capsize=3)

ax.set_xticks(x, ["Orientation in 8ms", "Orientation in 80ms", "Orientation in 400ms"])
# ax.set_xticks(x, included_targets)
ax.legend(ncol=3)

ax.set_ylabel("Predictability (R2)")

ax.set_ylim(0, 1)

# plt.show()

# set figure size to 8x2 inches
fig.set_size_inches(8, 2)

# save the figure to a pdf file without whitespace around it
plt.savefig(f"predictability_translation_grouped.pdf", format="pdf", bbox_inches="tight")



from typing import Optional

import mujoco
import numpy as np
from gymnasium import utils
from scipy.spatial import transform

import angorapy.tasks.utils
from angorapy.common.const import N_SUBSTEPS
from angorapy.common.senses import Sensation
from angorapy.tasks.envs.dexterity.consts import DEFAULT_INITIAL_QPOS
from angorapy.tasks.envs.dexterity.consts import FINGERTIP_SITE_NAMES
from angorapy.tasks.envs.dexterity.core import BaseShadowHandEnv
from angorapy.tasks.envs.dexterity.mujoco_model.worlds.manipulation import ShadowHandWithCubeWorld
from angorapy.tasks.envs.dexterity.reward import manipulate
from angorapy.tasks.envs.dexterity.utils import quat_from_angle_and_axis
from angorapy.tasks.envs.dexterity.reward_configs import MANIPULATE_BASE
from angorapy.tasks.utils import robot_get_obs


class BaseManipulate(BaseShadowHandEnv):
    """Base class for in-hand object manipulation tasks."""

    max_steps_per_goal = 100

    def __init__(
            self,
            target_position,
            target_rotation,
            target_position_range,
            initial_qpos=DEFAULT_INITIAL_QPOS,
            randomize_initial_position=True,
            randomize_initial_rotation=True,
            distance_threshold=0.01,
            rotation_threshold=0.4,
            relative_control=False,
            ignore_z_target_rotation=False,
            n_substeps=N_SUBSTEPS,
            delta_t=0.002,
            touch_get_obs="sensordata",
            vision=False,
            touch=True,
            render_mode: Optional[str] = None
    ):
        """Initializes a new Hand manipulation environment.

        Args:
            target_position (string): the type of target position:
                - ignore: target position is fully ignored, i.e. the object can be positioned arbitrarily
                - fixed: target position is set to the initial position of the object
                - random: target position is fully randomized according to target_position_range
            target_rotation (string): the type of target rotation:
                - ignore: target rotation is fully ignored, i.e. the object can be rotated arbitrarily
                - fixed: target rotation is set to the initial rotation of the object
                - xyz: fully randomized target rotation around the X, Y and Z axis
                - z: fully randomized target rotation around the Z axis
                - parallel: fully randomized target rotation around Z and axis-aligned rotation around X, Y
            ignore_z_target_rotation (boolean): whether or not the Z axis of the target rotation is ignored
            target_position_range (np.array of shape (3, 2)): range of the target_position randomization
            initial_qpos (dict): a dictionary of joint names and values that define the initial configuration
            randomize_initial_position (boolean): whether or not to randomize the initial position of the object
            randomize_initial_rotation (boolean): whether or not to randomize the initial rotation of the object
            distance_threshold (float, in meters): the threshold after which the position of a goal is considered
            achieved
            rotation_threshold (float, in radians): the threshold after which the rotation of a goal is considered
            achieved
            n_substeps (int): number of substeps the simulation runs on every call to step
            relative_control (boolean): whether or not the hand is actuated in absolute joint positions or relative
            to the current state
            touch_get_obs (string): touch sensor readings
                - "boolean": returns 1 if touch sensor reading != 0.0 else 0
                - "sensordata": returns original touch sensor readings from self.sim.data.sensordata[id]
                - "log": returns log(x+1) touch sensor readings from self.sim.data.sensordata[id]
                - "off" or else: does not add touch sensor readings to the observation
            vision (bool): indicator whether the environment should return frames (True) or the exact object
                position (False)
        """

        self.world = ShadowHandWithCubeWorld()
        self.object_id = self.world.cube.name + "/"
        self.object_joint_id = f'{self.object_id}object:joint/'
        self.object_center_id = f'{self.object_id}object:center'

        self.touch_get_obs = touch_get_obs
        self.vision = vision
        self.touch = touch
        self.touch_color = [1, 0, 0, 0.5]
        self.notouch_color = [0, 0.5, 0, 0.2]

        self.target_position = target_position
        self.target_rotation = target_rotation
        self.target_position_range = target_position_range
        self.parallel_quats = [angorapy.tasks.utils.euler2quat(r) for r in
                               angorapy.tasks.utils.get_parallel_rotations()]
        self.randomize_initial_rotation = randomize_initial_rotation
        self.randomize_initial_position = randomize_initial_position
        self.distance_threshold = distance_threshold
        self.rotation_threshold = rotation_threshold
        self.ignore_z_target_rotation = ignore_z_target_rotation

        assert self.target_position in ['ignore', 'fixed', 'random']
        assert self.target_rotation in ['ignore', 'fixed', 'xyz', 'z', 'parallel']

        self.consecutive_goals_reached = 0
        self.steps_with_current_goal = 0
        self.previous_object_pose = self.get_object_pose()

        super().__init__(
            initial_qpos=initial_qpos,
            n_substeps=n_substeps,
            delta_t=delta_t,
            relative_control=relative_control,
            model=self.world.root,
            vision=vision,
            touch=touch,
            render_mode=render_mode
        )

        # set touch sensors rgba values
        for _, site_id in self._touch_sensor_id_site_id:
            self.model.site_rgba[site_id][3] = 0.0

    def _set_default_reward_function_and_config(self):
        self.reward_function = manipulate
        self.reward_config = MANIPULATE_BASE

    def assert_reward_setup(self):
        """Assert whether the reward config fits the environment. """
        assert set(MANIPULATE_BASE.keys()).issubset(
            self.reward_config.keys()), "Incomplete manipulate reward configuration."

    def get_object_pose(self):
        """Object position and rotation."""
        if hasattr(self, "data"):
            object_qpos = self.data.jnt(self.object_joint_id).qpos
            assert object_qpos.shape == (7,)
            return object_qpos.copy()
        else:
            # simulation not initialized yet
            return np.zeros(7)  # todo double check

    def _goal_distance(self,
                       goal_a,
                       goal_b):
        assert goal_a.shape == goal_b.shape
        assert goal_a.shape[-1] == 7

        d_pos = np.zeros_like(goal_a[..., 0])
        d_rot = np.zeros_like(goal_b[..., 0])

        if self.target_position != 'ignore':
            delta_pos = goal_a[..., :3] - goal_b[..., :3]
            d_pos = np.linalg.norm(delta_pos, axis=-1)

        if self.target_rotation != 'ignore':
            quat_a, quat_b = goal_a[..., 3:], goal_b[..., 3:]

            if self.ignore_z_target_rotation:
                # Special case: We want to ignore the Z component of the rotation.
                # This code here assumes Euler angles with xyz convention. We first transform
                # to euler, then set the Z component to be equal between the two, and finally
                # transform back into quaternions.
                euler_a = angorapy.tasks.utils.quat2euler(quat_a)
                euler_b = angorapy.tasks.utils.quat2euler(quat_b)
                euler_a[2] = euler_b[2]
                quat_a = angorapy.tasks.utils.euler2quat(euler_a)

            # Subtract quaternions and extract angle between them.
            quat_diff = angorapy.tasks.utils.quat_mul(quat_a, angorapy.tasks.utils.quat_conjugate(quat_b))
            angle_diff = 2 * np.arccos(np.clip(quat_diff[..., 0], -1., 1.))
            d_rot = angle_diff

        assert d_pos.shape == d_rot.shape
        return d_pos, d_rot

    def _is_success(self,
                    achieved_goal,
                    desired_goal):
        d_pos, d_rot = self._goal_distance(achieved_goal, desired_goal)
        achieved_pos = (d_pos < self.distance_threshold).astype(np.float32)
        achieved_rot = (d_rot < self.rotation_threshold).astype(np.float32)
        achieved_both = achieved_pos * achieved_rot

        return achieved_both

    def _env_setup(self,
                   initial_state):
        super()._env_setup(initial_state)

        self.initial_goal = self.get_object_pose().copy()
        self.goal = self._sample_goal()

    def reset(self,
              **kwargs):
        self.consecutive_goals_reached = 0
        self.steps_with_current_goal = 0

        return super().reset(**kwargs)

    def _reset_sim(self):
        self.reset_model()

        initial_qpos = self.data.jnt(self.object_joint_id).qpos.copy()
        initial_pos, initial_quat = initial_qpos[:3], initial_qpos[3:]
        assert initial_qpos.shape == (7,)
        assert initial_pos.shape == (3,)
        assert initial_quat.shape == (4,)
        initial_qpos = None

        # Randomization initial rotation.
        if self.randomize_initial_rotation:
            if self.target_rotation == 'z':
                angle = self.np_random.uniform(-np.pi, np.pi)
                axis = np.array([0., 0., 1.])
                offset_quat = quat_from_angle_and_axis(angle, axis)
                initial_quat = angorapy.tasks.utils.quat_mul(initial_quat, offset_quat)
            elif self.target_rotation == 'parallel':
                angle = self.np_random.uniform(-np.pi, np.pi)
                axis = np.array([0., 0., 1.])
                z_quat = quat_from_angle_and_axis(angle, axis)
                parallel_quat = self.parallel_quats[self.np_random.randint(len(self.parallel_quats))]
                offset_quat = angorapy.tasks.utils.quat_mul(z_quat, parallel_quat)
                initial_quat = angorapy.tasks.utils.quat_mul(initial_quat, offset_quat)
            elif self.target_rotation in ['xyz', 'ignore']:
                angle = self.np_random.uniform(-np.pi, np.pi)
                axis = self.np_random.uniform(-1., 1., size=3)
                offset_quat = quat_from_angle_and_axis(angle, axis)
                initial_quat = angorapy.tasks.utils.quat_mul(initial_quat, offset_quat)
            elif self.target_rotation == 'fixed':
                pass
            else:
                raise ValueError('Unknown target_rotation option "{}".'.format(self.target_rotation))

        # Randomize initial position.
        if self.randomize_initial_position:
            if self.target_position != 'fixed':
                initial_pos += self.np_random.normal(size=3, scale=0.005)

        initial_quat /= np.linalg.norm(initial_quat)
        initial_qpos = np.concatenate([initial_pos, initial_quat])
        self.data.jnt(self.object_joint_id).qpos[:] = initial_qpos

        def is_on_palm():
            mujoco.mj_forward(self.model, self.data)
            cube_middle_pos = self.data.site(self.object_center_id).xpos
            is_on_palm = (cube_middle_pos[2] > 0.04)
            return is_on_palm

        # Run the simulation for a bunch of timesteps to let everything settle in.
        for _ in range(10):
            self._set_action(self.data.actuator_length[:])
            mujoco.mj_step(self.model, self.data)

        return is_on_palm()

    def _sample_goal(self):
        # Select a goal for the object position.
        target_pos = None
        if self.target_position == 'random':
            assert self.target_position_range.shape == (3, 2)
            offset = self.np_random.uniform(self.target_position_range[:, 0], self.target_position_range[:, 1])
            assert offset.shape == (3,)
            target_pos = self.data.jnt(self.object_joint_id).qpos[:3] + offset
        elif self.target_position in ['ignore', 'fixed']:
            target_pos = self.data.jnt(self.object_joint_id).qpos[:3]
        else:
            raise ValueError('Unknown target_position option "{}".'.format(self.target_position))
        assert target_pos is not None
        assert target_pos.shape == (3,)

        # Select a goal for the object rotation.
        target_quat = None
        if self.target_rotation == 'z':
            angle = self.np_random.uniform(-np.pi, np.pi)
            axis = np.array([0., 0., 1.])
            target_quat = quat_from_angle_and_axis(angle, axis)
        elif self.target_rotation == 'parallel':
            angle = self.np_random.uniform(-np.pi, np.pi)
            axis = np.array([0., 0., 1.])
            target_quat = quat_from_angle_and_axis(angle, axis)
            parallel_quat = self.parallel_quats[self.np_random.randint(len(self.parallel_quats))]
            target_quat = angorapy.tasks.utils.quat_mul(target_quat, parallel_quat)
        elif self.target_rotation == 'xyz':
            angle = self.np_random.uniform(-np.pi, np.pi)
            axis = self.np_random.uniform(-1., 1., size=3)
            target_quat = quat_from_angle_and_axis(angle, axis)
        elif self.target_rotation in ['ignore', 'fixed']:
            target_quat = self.data.jnt(self.object_joint_id).qpos
        else:
            raise ValueError('Unknown target_rotation option "{}".'.format(self.target_rotation))
        assert target_quat is not None
        assert target_quat.shape == (4,)

        target_quat /= np.linalg.norm(target_quat)  # normalized quaternion
        goal = np.concatenate([target_pos, target_quat])
        return goal

    def _render_callback(self,
                         render_targets=False):
        # Assign current state to target object but offset a bit so that the actual object
        # is not obscured.

        if render_targets:
            goal = self.goal.copy()
            assert goal.shape == (7,)
            if self.target_position == 'ignore':
                # Move the object to the side since we do not care about it's position.
                goal[0] += 0.15

            self.data.jnt('target:joint').qpos[:] = goal
            self.data.jnt('target:joint').qvel[:] = np.zeros(6)

            if 'object_hidden' in self.model.names.split(b"\x00"):
                hidden_id = self.model.geom_name2id('object_hidden')
                self.model.geom_rgba[hidden_id, 3] = 1.

        mujoco.mj_forward(self.model, self.data)

    def _get_obs(self):

        """Gather humanoid senses and asymmetric information."""

        object_qpos = self.data.jnt(self.object_joint_id).qpos.copy()

        # vision
        if not self.vision:
            vision_input = object_qpos.astype(np.float32)
        else:
            tmp_render_mode = self.render_mode
            self.render_mode = "rgb_array"
            vision_input = self.render()
            self.render_mode = tmp_render_mode

        # goal
        target_orientation = self.goal.ravel().copy()[3:]

        # proprioception
        hand_joint_angles, hand_joint_velocities = robot_get_obs(self.model, self.data)

        proprioception = np.concatenate([
            hand_joint_angles,
            hand_joint_velocities
        ])

        # touch
        touch = self.get_touch()

        # asymmetric information
        finger_tip_positions = np.array([self.data.site(name).xpos for name in FINGERTIP_SITE_NAMES]).flatten()
        target_quat = transform.Rotation.from_quat(target_orientation)
        current_quat = transform.Rotation.from_quat((object_qpos[3:]))
        relative_target_orientation = (target_quat * current_quat.inv()).as_quat()

        object_positional_velocity = self.data.body(self.object_id).cvel[3:]  # todo double check if correct part
        object_angular_velocity = self.data.body(self.object_id).cvel[:3]  # quaternions in OpenAI model

        asymmetric = np.concatenate([
            finger_tip_positions,
            relative_target_orientation,
            object_positional_velocity,
            object_angular_velocity
        ])

        return {
            "observation": Sensation(
                vision=vision_input,
                proprioception=proprioception.copy(),
                touch=touch,
                goal=target_orientation,
                asymmetric=None if not self.asymmetric else asymmetric
            ),
        }

    def _is_dropped(self) -> bool:
        """Heuristically determine whether the object still is in the hand."""

        # determine object center position
        obj_center_pos = self.data.site(self.object_center_id).xpos

        # determine palm center position
        # palm_center_pos = self.data.site("robot/palm_center_site").xpos
        palm_center_pos = self.data.site("robot/palm_center_site").xpos

        dropped = (
                obj_center_pos[2] < palm_center_pos[2]  # z axis of object smaller than that of palm
        )

        return dropped

    def _goal_progress(self):
        return (sum(self._goal_distance(self.goal, self.previous_object_pose))
                - sum(self._goal_distance(self.goal, self.get_object_pose())))

    def step(self,
             action):
        """Make step in environment."""
        obs, reward, terminated, truncated, info = super().step(action)
        self.steps_with_current_goal += 1

        success = self._is_success(self.get_object_pose(), self.goal)

        # determine if a goal has been reached
        if not success:
            self.previous_object_pose = self.get_object_pose().copy()
        else:
            self.consecutive_goals_reached += 1
            self.goal = self._sample_goal()
            self.steps_with_current_goal = 0
            obs = self._get_obs()

        if self.steps_with_current_goal >= BaseManipulate.max_steps_per_goal:
            terminated = True

        # determine if done
        dropped = self._is_dropped()
        terminated = terminated or dropped or self.consecutive_goals_reached >= 50

        if "auxiliary_performances" not in info.keys():
            info["auxiliary_performances"] = {}
        info["auxiliary_performances"]["consecutive_goals_reached"] = self.consecutive_goals_reached

        return obs, reward, terminated, truncated, info

    def _get_info(self):
        return {
            **super()._get_info(),
            "is_success": self._is_success(self.goal, self.get_object_pose()),
            "achieved_goal": self.get_object_pose().copy(),
            "desired_goal": self.goal.copy(),
        }


class ManipulateBlock(BaseManipulate, utils.EzPickle):
    """Manipulate Environment with a Block as an object."""

    asymmetric = False
    continuous = True

    def __init__(self,
                 target_position='ignore',
                 target_rotation='xyz',
                 touch_get_obs='sensordata',
                 relative_control=True,
                 vision: bool = False,
                 delta_t: float = 0.002,
                 render_mode: Optional[str] = None):
        utils.EzPickle.__init__(self, target_position, target_rotation, touch_get_obs, "dense")
        BaseManipulate.__init__(self,
                                touch_get_obs=touch_get_obs,
                                target_rotation=target_rotation,
                                target_position=target_position,
                                target_position_range=np.array([(-0.04, 0.04), (-0.06, 0.02), (0.0, 0.06)]),
                                vision=vision,
                                relative_control=relative_control,
                                delta_t=delta_t,
                                render_mode=render_mode
                                )


class ManipulateBlockDiscrete(ManipulateBlock):
    asymmetric = False
    continuous = False


class ManipulateBlockDiscreteAsymmetric(ManipulateBlockDiscrete):
    asymmetric = True
    continuous = False


class ManipulateBlockAsymmetric(ManipulateBlockDiscrete):
    asymmetric = True
    continuous = True



import argparse
import json

import matplotlib.pyplot as plt
import numpy as np
import scipy.stats as st

from dexterity.design.common import QUALITATIVE_COLOR_PALETTE
from utilities.plotting import plot_with_confidence

parser = argparse.ArgumentParser()
parser.add_argument("exp", nargs="*", type=str, default=["default_Acrobot-v1"],
                    help="name of the experiment")
parser.add_argument("--show", action="store_true")
parser.add_argument("--ignore", nargs="*", type=str, default=[], help="list of ignored configs")
parser.add_argument("--only", nargs="*", type=str, default=[], help="list of included configs")
parser.add_argument("--no-confidence", action="store_true")
parser.add_argument("--cycles", type=int)

args = parser.parse_args()

benchmark_data = {}
for exp in args.exp:
    with open(f"docs/benchmarks/{exp}.json") as f:
        benchmark_data.update(json.load(f))

results = benchmark_data["results"]
meta = benchmark_data["meta"]

plt.axhline(meta["reward_threshold"], ls="--", color="grey") if meta["reward_threshold"] is not None else None

x = list(range(1, len(results[list(results.keys())[0]]["means"]) + 1))

include_configs = results.keys()
if len(args.only) > 0:
    include_configs = [c for c in include_configs if c in args.only]
elif len(args.ignore) > 0:
    include_configs = [c for c in include_configs if c not in args.ignore]

for i, name in enumerate(include_configs):

    sub_exp = results[name]

    means = sub_exp["means"]
    stdevs = np.sqrt(sub_exp["var"])
    n = sub_exp["n"]
    h = st.t.interval(0.95, sub_exp["n"], loc=means, scale=stdevs / np.sqrt(n))

    plot_with_confidence(x=means[:args.cycles],
                         lb=h[0][:args.cycles],
                         ub=h[1][:args.cycles],
                         label=f"{name} [{sub_exp['n']}]",
                         col=QUALITATIVE_COLOR_PALETTE[i],
                         alpha=0 if args.no_confidence else 0.2)

    ax = plt.gca()
    # ax.set_facecolor("#FBFBFB")

    plt.xlabel("Cycle")
    plt.ylabel("Mean Cumulative Reward")

plt.legend(loc="lower right")
plt.title(f"{args.exp[0].split('_')[1]}")
plt.savefig(f"docs/benchmarks/benchmarking-{'-'.join(include_configs)}-{'-'.join(args.exp)}.pdf", format="pdf")
if args.show:
    plt.show()



#!/usr/bin/env python
import argparse

from angorapy.agent.ppo_agent import PPOAgent
from angorapy.analysis import investigators

parser = argparse.ArgumentParser(description="Inspect the dynamics an episode of an agent.")
parser.add_argument("id", type=int, nargs="?", help="id of the agent, defaults to newest", default=1653053413)
parser.add_argument("--state", type=str, help="state, either iteration or 'best'", default="best")

args = parser.parse_args()
args.state = int(args.state) if args.state not in ["b", "best", "last"] else args.state

agent = PPOAgent.from_agent_state(args.id, args.state, path_modifier="../../")
investigator = investigators.LatentDynamics.from_agent(agent)
env = agent.env

investigator.prepare(env, layer="pmc_recurrent_layer", n_states=25 * 1024)
investigator.fit(batch_size=1024)



import os
import sys
from typing import List

import numpy as np
import tensorflow as tf
import tqdm
from angorapy.utilities.core import flatten
from angorapy.utilities.core import stack_dicts
from matplotlib import pyplot as plt

from angorapy.analysis.investigators import base_investigator
from angorapy.common.policies import BasePolicyDistribution
from angorapy.tasks.wrappers import TaskWrapper


class ActivityAutocorrelation(base_investigator.Investigator):
    """Investigate autocorrelation of the activity in different regions of the network as a measure of how long
    information is maintained over time."""

    def __init__(self, network: tf.keras.Model, distribution: BasePolicyDistribution):
        super().__init__(network, distribution)
        self._data = None
        self.prepared = False

        self.n_states = np.nan

    def prepare(self, env: TaskWrapper, layers: List[str], n_states=1000, n_repeats=1, verbose=True):
        """Prepare samples to predict from."""
        self.network.reset_states()
        state, info = env.reset(return_info=True)

        repeated_collections = []
        for i in range(n_repeats):
            activation_collection = []
            for _ in tqdm.tqdm(range(n_states), disable=not verbose):
                prepared_state = state.with_leading_dims(time=self.is_recurrent).dict()

                # make step and record activities
                activations = self.get_layer_activations(layers, prepared_state)
                probabilities = flatten(activations["output"])
                activations.pop("output")
                activation_collection.append(activations)

                action, _ = self.distribution.act(*probabilities)

                observation, reward, terminated, truncated, info = env.step(action)
                done = terminated or truncated

                if done:
                    state, info = env.reset(return_info=True)
                    self.network.reset_states()
                else:
                    state = observation

            repeated_collections.append(activation_collection)

        self._data = stack_dicts([stack_dicts(ac) for ac in repeated_collections])
        self.prepared = True

    def fit(self, n_lags=30):
        """Measure the predictability of target_information based on the information in source_layer's activation."""
        assert self.prepared, "Need to prepare before investigating."

        # calculate autocorrelation for each layer
        autocorrelation = {}
        for layer in self._data.keys():
            layer_autocorrelations = []
            for repeat in range(self._data[layer].shape[0]):
                layer_autocorrelations.append(self.calculate_statewise_correlation(
                    np.squeeze(self._data[layer][repeat]), n_lags=n_lags)
                )

            autocorrelation[layer] = np.mean(layer_autocorrelations, axis=0)

        return autocorrelation

    def calculate_statewise_correlation(self, data, n_lags=30):
        corr_coeffs = np.corrcoef(data)
        lag_shifted_corr_coeffs = np.zeros_like(corr_coeffs)
        for row in range(n_lags):
            lag_shifted_corr_coeffs[row] = np.roll(corr_coeffs[row], -row)

        return lag_shifted_corr_coeffs[:n_lags, :n_lags]



import os

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

import tensorflow as tf

from angorapy import make_task

import matplotlib.pyplot as plt

tf.get_logger().setLevel('INFO')

env = make_task("ManipulateBlockVisualDiscrete-v0", render_mode="rgb_array")

state = env.reset()

for i in range(100):
    state, r, dd, truncated, info = env.step(env.action_space.sample())

plt.imshow(state.vision / 1)
plt.show()



#!/usr/bin/env python
"""Data reading and writing utilities for distributed learning."""
import os
import random
import re
from functools import partial
from typing import Union, Tuple, List

import numpy as np
import tensorflow as tf

from angorapy.common.const import STORAGE_DIR
from angorapy.common.senses import Sensation
from angorapy.utilities.datatypes import StatBundle
from angorapy.common.data_buffers import ExperienceBuffer, TimeSequenceExperienceBuffer


def _float_feature(value):
    """Returns a float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))


def _bytes_feature(value):
    """"Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy()
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def serialize_sample(*args, feature_names):
    """Serialize a multi-input (shadow hand) sample from a dataset."""
    assert len(args) == len(feature_names), "Sample serialization reveived an unequal number of features than names."

    feature = {}
    for i, feature_name in enumerate(feature_names):
        feature[feature_name] = _bytes_feature(tf.io.serialize_tensor(args[i]))

    # Create a Features message using tf.train.Example.
    example_proto = tf.train.Example(features=tf.train.Features(feature=feature))
    return example_proto.SerializeToString()


def tf_serialize_example(sample):
    """TF wrapper for serialization function."""
    feature_names = ([sense for sense in ["vision", "somatosensation", "proprioception", "goal"] if sense in sample]
                     + ["action", "action_prob", "return", "advantage", "value", "done", "mask"])

    inputs = [sample[f] for f in feature_names]

    tf_string = tf.py_function(partial(serialize_sample, feature_names=feature_names), inputs, tf.string)
    return tf.reshape(tf_string, ())


def make_dataset_and_stats(buffer: ExperienceBuffer) -> Tuple[tf.data.Dataset, StatBundle]:
    """Make dataset object and StatBundle from ExperienceBuffer."""
    completed_episodes = buffer.episodes_completed
    numb_processed_frames = buffer.capacity * buffer.seq_length

    # build the dataset
    tensor_slices = {
        "action": buffer.actions,
        "action_prob": buffer.action_probabilities,
        "return": buffer.returns,
        "advantage": buffer.advantages,
        "value": buffer.values,
        "done": buffer.dones,
        "mask": buffer.mask
    }

    tensor_slices.update(buffer.states)

    dataset = tf.data.TFRecordDataset.from_tensor_slices(tensor_slices)

    # make statistics object
    underflow = None
    if isinstance(buffer, TimeSequenceExperienceBuffer):
        underflow = round(1 - (buffer.true_number_of_transitions / (buffer.capacity * buffer.seq_length)), 2)

    stats = StatBundle(
        completed_episodes,
        numb_processed_frames,
        buffer.episode_rewards,
        buffer.episode_lengths,
        tbptt_underflow=underflow,
        per_receptor_mean={
            # mean over all buf the last dimension to mean per receptor (reducing batch and time dimensions)
            # todo ignore empty masked states
            sense: np.mean(buffer.states[sense], axis=tuple(range(len(buffer.states[sense].shape)))[:-1])
            for sense in buffer.states.keys()
        },
        auxiliary_performances=buffer.auxiliary_performances
    )

    return dataset, stats


def read_dataset_from_storage(dtype_actions: tf.dtypes.DType, id_prefix: Union[str, int], responsive_senses: List[str],
                              shuffle: bool = True, worker_ids: List = None):
    """Read all files in storage into a tf record dataset without actually loading everything into memory.

    Args:
        dtype_actions:      datatype of the actions
        id_prefix:          prefix (usually agent id) by which to filter available experience files
        responsive_senses:  list of senses (string represented, in ["proprioception", "vision", "somatosensation",
                            "goal"]) that the agent utilizes
        shuffle:            whether or not to shuffle the datafiles
    """
    assert all(r in Sensation.sense_names for r in responsive_senses)

    feature_description = {
        "action": tf.io.FixedLenFeature([], tf.string),
        "action_prob": tf.io.FixedLenFeature([], tf.string),
        "return": tf.io.FixedLenFeature([], tf.string),
        "advantage": tf.io.FixedLenFeature([], tf.string),
        "value": tf.io.FixedLenFeature([], tf.string),
        "done": tf.io.FixedLenFeature([], tf.string),
        "mask": tf.io.FixedLenFeature([], tf.string),

        # STATES
        **{sense: tf.io.FixedLenFeature([], tf.string) for sense in responsive_senses}
    }

    def _parse_function(example_proto):
        # Parse the input `tf.Example` proto using the dictionary above.
        parsed = tf.io.parse_single_example(example_proto, feature_description)

        parsed["action"] = tf.io.parse_tensor(parsed["action"], out_type=dtype_actions)
        parsed["action_prob"] = tf.io.parse_tensor(parsed["action_prob"], out_type=tf.float32)
        parsed["return"] = tf.io.parse_tensor(parsed["return"], out_type=tf.float32)
        parsed["advantage"] = tf.io.parse_tensor(parsed["advantage"], out_type=tf.float32)
        parsed["value"] = tf.io.parse_tensor(parsed["value"], out_type=tf.float32)
        parsed["done"] = tf.io.parse_tensor(parsed["done"], out_type=tf.bool)
        parsed["mask"] = tf.io.parse_tensor(parsed["mask"], out_type=tf.bool)

        for sense in responsive_senses:
            parsed[sense] = tf.io.parse_tensor(parsed[sense], out_type=tf.float32)

        return parsed

    worker_id_regex = "[0-9]*"
    if worker_ids is not None:
        worker_id_regex = "(" + "|".join(map(str, worker_ids)) + ")"

    files = [os.path.join(STORAGE_DIR, name) for name in os.listdir(STORAGE_DIR) if
             re.match(f"{id_prefix}_data_{worker_id_regex}\.tfrecord", name)]

    random.shuffle(files) if shuffle else None

    serialized_dataset = tf.data.TFRecordDataset(files)
    serialized_dataset = serialized_dataset.map(_parse_function)

    return serialized_dataset



import random
from typing import Optional

import mujoco
import numpy as np

from angorapy.common.const import N_SUBSTEPS
from angorapy.tasks.envs.dexterity.consts import DEFAULT_INITIAL_QPOS
from angorapy.tasks.envs.dexterity.consts import FINGERTIP_SITE_NAMES
from angorapy.tasks.envs.dexterity.core import BaseShadowHandEnv
from angorapy.tasks.envs.dexterity.mujoco_model.worlds.reach import ShadowHandReachWorld
from angorapy.tasks.envs.dexterity.reward import free_reach
from angorapy.tasks.envs.dexterity.reward import reach
from angorapy.tasks.envs.dexterity.reward import sequential_free_reach
from angorapy.tasks.envs.dexterity.reward import sequential_reach
from angorapy.tasks.envs.dexterity.utils import generate_random_sim_qpos
from angorapy.tasks.envs.dexterity.utils import get_fingertip_distance
from angorapy.tasks.envs.dexterity.reward_configs import REACH_BASE
from angorapy.tasks.utils import mj_qpos_dict_to_qpos_vector


class Reach(BaseShadowHandEnv):
    """Simple Reaching task."""

    def __init__(self,
                 initial_qpos=DEFAULT_INITIAL_QPOS,
                 n_substeps=N_SUBSTEPS,
                 relative_control=True,
                 vision=False,
                 touch=True,
                 force_finger=None,
                 render_mode: Optional[str] = None):
        assert force_finger in list(range(5)) + [None], "Forced finger index out of range [0, 5]."

        self.forced_finger = force_finger
        self.current_target_finger = self.forced_finger

        # STATE INITIALIZATION
        assert initial_qpos in ["random", "buffered"] or isinstance(initial_qpos, dict), "Illegal state initialization."

        self.state_initialization = initial_qpos
        if self.state_initialization == "random":
            initial_qpos = generate_random_sim_qpos(DEFAULT_INITIAL_QPOS)
        elif self.state_initialization == "buffered":
            initial_qpos = DEFAULT_INITIAL_QPOS

        self.world = ShadowHandReachWorld()

        super().__init__(
            initial_qpos=initial_qpos,
            n_substeps=n_substeps,
            relative_control=relative_control,
            vision=vision,
            touch=touch,
            render_mode=render_mode,
            model=self.world.root,
        )

        self.previous_finger_positions = [self.get_finger_position(fname) for fname in FINGERTIP_SITE_NAMES]

    def _set_default_reward_function_and_config(self):
        self.reward_function = reach
        self.reward_config = REACH_BASE

    def assert_reward_setup(self):
        """Assert whether the reward config fits the environment. """
        assert set(REACH_BASE.keys()).issubset(self.reward_config.keys()), "Incomplete free reach reward configuration."

    def _get_achieved_goal(self):
        return self.get_fingertip_positions()

    def _is_success(self,
                    achieved_goal,
                    desired_goal):
        d = get_fingertip_distance(achieved_goal, desired_goal)
        return (d < self.reward_config["SUCCESS_DISTANCE"]).astype(np.float32)

    def get_thumbs_previous_position(self) -> np.ndarray:
        """Get the previous position of the thumb in space."""
        return self.previous_finger_positions[-1]

    def get_fingers_previous_position(self,
                                      fname) -> np.ndarray:
        """Get the previous position of the given finger in space."""
        return self.previous_finger_positions[FINGERTIP_SITE_NAMES.index(fname)]

    def get_target_finger_position(self) -> np.ndarray:
        """Get position of the target finger in space."""
        return self.data.site(FINGERTIP_SITE_NAMES[self.current_target_finger]).xpos.flatten()

    def get_target_fingers_previous_position(self) -> np.ndarray:
        """Get position of the target finger in space."""
        return self.previous_finger_positions[self.current_target_finger]

    def _reset_sim(self):
        """Resets a simulation and indicates whether or not it was successful."""
        self.set_state(**self.initial_state)  # reset everything

        if self.state_initialization == "random":
            initial_qpos = generate_random_sim_qpos(DEFAULT_INITIAL_QPOS)

            self.set_state(
                qpos=mj_qpos_dict_to_qpos_vector(self.model, initial_qpos),
                qvel=self.initial_state["qvel"]
            )
        # elif self.state_initialization == "buffered":
        #     # pull state from buffer as initial state for the environment
        #     if len(self.state_memory_buffer) > 100:  # TODO as variable in config?
        #         sampled_initial_state = random.choice(self.state_memory_buffer)
        #         self.sim.set_state(sampled_initial_state)

        return True

    def _sample_goal(self):
        finger_names = [name for name in FINGERTIP_SITE_NAMES if name != self.thumb_name]

        if self.forced_finger is None:
            finger_name = self.np_random.choice(finger_names)
        else:
            finger_name = finger_names[self.forced_finger]

        thumb_idx = FINGERTIP_SITE_NAMES.index(self.thumb_name)
        finger_idx = FINGERTIP_SITE_NAMES.index(finger_name)
        self.current_target_finger = finger_idx
        goal = self.initial_goal.copy().reshape(-1, 3)

        assert thumb_idx != finger_idx

        # Pick a meeting point above the hand.
        meeting_pos = self.palm_xpos + np.array([0.08, 0.01, 0.05])

        # move meeting point along y-axis towards the target finger to make goal achievable
        meeting_pos[1] -= .5 * (meeting_pos - goal[finger_idx])[1]

        # perturb meeting point
        meeting_pos += self.np_random.normal(scale=0.005, size=meeting_pos.shape)

        # slightly move meeting goal towards the respective finger to avoid that they overlap
        for idx in [thumb_idx, finger_idx]:
            offset_direction = (meeting_pos - goal[idx])
            offset_direction /= np.linalg.norm(offset_direction)
            goal[idx] = meeting_pos - 0.005 * offset_direction

        # in 10% of the cases, the goal is the initial goal (i.e. the hand is not supposed to move)
        if self.forced_finger is None:
            if self.np_random.uniform() < 0.1:
                goal = self.initial_goal.copy()

        return goal.flatten()

    def _env_setup(self, initial_state):
        super()._env_setup(initial_state=initial_state)

        if initial_state is not None:
            self.set_state(**initial_state)

        self.initial_goal = self._get_achieved_goal().copy()
        self.palm_xpos = self.data.body(self.palm_name).xpos.copy()

        self.goal = self._sample_goal()

    def _render_callback(self, render_targets=True):
        if render_targets:
            # Visualize targets.
            sites_offset = (self.data.site_xpos - self.model.site_pos).copy()
            goal = self.goal.reshape(5, 3)
            for finger_idx in range(5):
                site_name = f"robot/target{finger_idx}"

                site_id = mujoco.mj_name2id(self.model, mujoco.mjtObj.mjOBJ_SITE, site_name)
                self.model.site(site_name).rgba[-1] = 0.2
                # self.data.site(site_name).xpos[:] = (goal[finger_idx] - sites_offset[site_id]).copy()
                self.model.site_pos[site_id] = (goal[finger_idx] - sites_offset[site_id]).copy()

            # Visualize finger positions.
            achieved_goal = self._get_achieved_goal().reshape(5, 3)
            for finger_idx in range(5):
                site_name = f"robot/finger{finger_idx}"
                site_id = mujoco.mj_name2id(self.model, mujoco.mjtObj.mjOBJ_SITE, site_name)
                self.model.site(site_name).rgba[-1] = 0.2
                # self.data.site(site_name).xpos[:] = (achieved_goal[finger_idx] - sites_offset[site_id]).copy()
                self.model.site_pos[site_id] = (achieved_goal[finger_idx] - sites_offset[site_id]).copy()

        mujoco.mj_forward(self.model, self.data)

    def step(self, action):
        """Step the environment."""
        self.previous_finger_positions = [self.get_finger_position(fname).copy() for fname in FINGERTIP_SITE_NAMES]

        return super().step(action)

    def _get_info(self):
        return {
            **super()._get_info(),
            "target_finger": self.current_target_finger,
            "achieved_goal": self._get_achieved_goal().copy(),
            "desired_goal": self.goal.copy(),
            "is_success": self._is_success(self._get_achieved_goal(), self.goal),
        }

    def _get_obs(self):
        """Return the observation."""
        obs = super()._get_obs()
        obs["observation"]["vision"] = self.get_fingertip_positions().flatten()

        return obs


class ReachSequential(Reach):
    """Generate finger configurations in a sequence. Each new goal is randomly generated as the old goal is reached."""

    def __init__(self,
                 initial_qpos=DEFAULT_INITIAL_QPOS,
                 n_substeps=N_SUBSTEPS,
                 relative_control=True,
                 vision=False,
                 touch=True):
        self.current_target_finger = None
        self.goal_sequence = []

        super().__init__(
            initial_qpos=initial_qpos,
            n_substeps=n_substeps,
            relative_control=relative_control,
            vision=vision,
            touch=touch)

    def _set_default_reward_function_and_config(self):
        self.reward_function = sequential_reach
        self.reward_config = REACH_BASE

    def _sample_goal(self):
        available_fingers = [0, 1, 2, 3]  # where None refers to not meeting any fingers
        if self.current_target_finger is not None:
            available_fingers.remove(self.current_target_finger)
        self.current_target_finger = random.choice(available_fingers)

        goal = self.initial_goal.copy()

        # pick a meeting point above the hand
        meeting_pos = self.palm_xpos + np.array([0.0, -0.2, 0.05])
        meeting_pos += self.np_random.normal(scale=0.005, size=meeting_pos.shape)

        # slightly move meeting goal towards the respective finger to avoid that they overlap
        if self.current_target_finger is not None:
            goal = goal.reshape(-1, 3)
            for idx in [self.current_target_finger, -1]:
                offset_direction = (meeting_pos - goal[idx])
                offset_direction /= np.linalg.norm(offset_direction)
                goal[idx] = meeting_pos - 0.005 * offset_direction

        # remember the goals
        self.goal_sequence.append(self.current_target_finger)

        return goal.flatten()

    def get_target_finger_position(self) -> np.ndarray:
        return self.sim.data.get_site_xpos(FINGERTIP_SITE_NAMES[self.current_target_finger]).flatten()

    def _is_success(self,
                    achieved_goal,
                    desired_goal):
        d = get_fingertip_distance(achieved_goal, desired_goal)
        return (d < self.reward_config["SUCCESS_DISTANCE"]).astype(np.float32)

    def reset(self):
        self.current_target_finger = None
        self.goal_sequence = []

        ret = super().reset()

        return ret, {}

    def step(self,
             action):
        observation, reward, done, truncated, info = super().step(action)

        # set next subgoal if current one is achieved
        if info["is_success"]:
            # print(
            #     f"Reached Target {self.current_target_finger} (after reaching {len(self.goal_sequence) - 1}
            #     targets)!")
            self.goal = self._sample_goal()

        return observation, reward, done, truncated, info


class FreeReach(Reach):
    """Reaching task where the actual position of the joint fingers is irrelevant.

    It only matters which fingertips need to be joint. The reward is based on the distance between the fingertips,
    punishing distance of the thumb to target fingers and rewarding the distance to non-target fingers.

    The goal is represented as a one-hot vector of size 4."""

    def _set_default_reward_function_and_config(self):
        self.reward_function = free_reach
        self.reward_config = REACH_BASE

    def _sample_goal(self):
        if self.forced_finger is None:
            finger_names = [name for name in FINGERTIP_SITE_NAMES if name != self.thumb_name]

            # choose the finger to join with the thumb
            finger_name = self.np_random.choice(a=finger_names, size=1, replace=False)

            # get finger id
            f_id = FINGERTIP_SITE_NAMES.index(finger_name)
        else:
            f_id = self.forced_finger

        self.current_target_finger = f_id

        # make one hot encoding
        goal = np.zeros(len(FINGERTIP_SITE_NAMES))
        goal[self.current_target_finger] = 1

        return goal

    def get_target_finger_position(self):
        """Get position of the target finger in space."""
        return self.data.site(FINGERTIP_SITE_NAMES[np.where(self.goal == 1)[0].item()]).xpos.flatten()

    def _is_success(self,
                    achieved_goal,
                    desired_goal):
        d = get_fingertip_distance(self.get_thumb_position(), self.get_target_finger_position())
        return (d < self.reward_config["SUCCESS_DISTANCE"]).astype(np.float32)

    def _render_callback(self,
                         render_targets=False):
        sites_offset = (self.data.site_xpos - self.model.site_pos).copy()

        # Visualize finger positions.
        achieved_goal = self._get_achieved_goal().reshape(5, 3)
        for finger_idx in range(5):
            site_name = f"robot/finger{finger_idx}"
            site_id = mujoco.mj_name2id(self.model, mujoco.mjtObj.mjOBJ_SITE, site_name)

            fname = FINGERTIP_SITE_NAMES[finger_idx]
            if not (fname == self.thumb_name or finger_idx == np.where(self.goal == 1)[0].item()):
                self.model.site_rgba[site_id][-1] = 0
                continue

            self.model.site(site_name).rgba[-1] = 0.2

            self.model.site_pos[site_id] = (
                    achieved_goal[finger_idx] - sites_offset[site_id]
            )
        mujoco.mj_forward(self.model, self.data)


class FreeReachSequential(FreeReach):
    """Freely join fingers in a sequence, where each new goal is randomly generated as the old goal is reached."""

    def __init__(self,
                 n_substeps=N_SUBSTEPS,
                 relative_control=True,
                 initial_qpos=DEFAULT_INITIAL_QPOS,
                 vision=False):
        self.goal_sequence = []

        self.force_exemplary_sequence = True
        self.default_exemplary_sequence = [0, 1, 2, 3]
        self.exemplary_sequence = self.default_exemplary_sequence.copy()

        super().__init__(n_substeps, relative_control, initial_qpos, vision)
        self.exemplary_sequence = self.default_exemplary_sequence.copy()  # reset after initial init stuff

    def _set_default_reward_function_and_config(self):
        self.reward_function = sequential_free_reach
        self.reward_config = REACH_BASE

    def _sample_goal(self):
        if not self.force_exemplary_sequence:
            available_fingers = [0, 1, 2, 3]
            available_fingers.remove(self.current_target_finger)
            self.current_target_finger = random.choice(available_fingers)
        else:
            self.current_target_finger = self.exemplary_sequence.pop(0)
            self.goal_sequence.append(self.current_target_finger)
            self.exemplary_sequence.append(self.current_target_finger)

        # make one hot encoding
        goal = np.zeros(len(FINGERTIP_SITE_NAMES))
        goal[self.current_target_finger] = 1

        self.goal_sequence.append(self.current_target_finger)

        return goal

    def get_target_finger_position(self) -> np.ndarray:
        return self.sim.data.get_site_xpos(FINGERTIP_SITE_NAMES[self.current_target_finger]).flatten()

    def _is_success(self,
                    achieved_goal,
                    desired_goal):
        d = get_fingertip_distance(self.get_thumb_position(),
                                   self.get_finger_position(FINGERTIP_SITE_NAMES[self.current_target_finger]))
        return (d < self.distance_threshold).astype(np.float32)

    def reset(self):
        self.goal_sequence = []
        self.exemplary_sequence = self.default_exemplary_sequence.copy()

        ret = super().reset()
        return ret, {}

    def _render_callback(self):
        sites_offset = (self.sim.data.site_xpos - self.sim.model.site_pos).copy()

        # Visualize finger positions.
        achieved_goal = self._get_achieved_goal().reshape(5, 3)
        for finger_idx in range(5):
            fname = FINGERTIP_SITE_NAMES[finger_idx]
            site_name = 'finger{}'.format(finger_idx)
            site_id = self.sim.model.site_name2id(site_name)
            if not (fname == self.thumb_name or finger_idx == self.current_target_finger):
                self.sim.model.site_rgba[site_id][-1] = 0
                continue

            self.sim.model.site_rgba[site_id][-1] = 0.2
            self.sim.model.site_pos[site_id] = achieved_goal[finger_idx] - sites_offset[site_id]

        self.sim.forward()

    def step(self,
             action):
        observation, reward, done, truncated, info = super().step(action)

        # set next subgoal if current one is achieved
        if info["is_success"]:
            print(
                f"Reached Target {self.current_target_finger} (after reaching {len(self.goal_sequence) - 1} targets)!")
            self.goal = self._sample_goal()

        return observation, reward, done, truncated, info



#!/usr/bin/env python
"""Constants used throughout the repository."""
import numpy as np

# MISC
from matplotlib.colors import ListedColormap
import seaborn as sns

COLORS = dict(
    HEADER='\033[95m',
    OKBLUE='\033[94m',
    OKGREEN='\033[92m',
    WARNING='\033[93m',
    FAIL='\033[91m',
    ENDC='\033[0m',
    BOLD='\033[1m',
    UNDERLINE='\033[4m',
    GREEN='\u001b[32m',
    ORANGE="\033[33m"
)

QUALITATIVE_COLOR_PALETTE = [
    "#1D6996",
    "#CC503E",
    "#73AF48",
    "#0F8554",
    "#E17C05",
    "#38A6A5",
    "#EDAD08",
    "#94346E",
    "#5F4690",
    "#6F4070",
    "#994E95",
    "#666666"
]

# MISC
MIN_STAT_EPS = 5
N_SUBSTEPS = 10
SHADOWHAND_MAX_STEPS = 100
SHADOWHAND_SEQUENCE_MAX_STEPS = 100

# STORAGE
BASE_SAVE_PATH = "storage/saved_models/"
STORAGE_DIR = "storage/experience/"
PRETRAINED_COMPONENTS_PATH = "storage/pretraining/"
PATH_TO_EXPERIMENTS = "storage/experiments/"
PATH_TO_BENCHMARKS = "docs/benchmarks/"

# NUMERICAL PRECISION
NP_FLOAT_PREC = np.float64
NUMPY_INTEGER_PRECISION = np.int64
EPSILON = 1e-6  # dont make this lower! 1e-8 would be ignored due to float32 precision

# SHAPES
VISION_WH = 128

# DEBUGGING
DETERMINISTIC = False
DEBUG = False

# OTHER STUFF
COLORMAP = list(ListedColormap(sns.color_palette("deep")).colors)
COLORMAP_HEX = ['#%02x%02x%02x' % tuple(int(c * 255) for c in color) for color in COLORMAP]



from typing import Union

import tensorflow as tf

from angorapy.agent.ppo import loss
from angorapy.common.senses import Sensation


@tf.function
def learn_on_batch(batch,
                   joint: tf.keras.Model,
                   distribution,
                   continuous_control: bool,
                   clip_values: bool,
                   clipping_bound: tf.Tensor,
                   gradient_clipping: Union[tf.Tensor, None],
                   c_value: tf.Tensor,
                   c_entropy: tf.Tensor,
                   is_recurrent: bool):
    """Optimize a given network on the given batch.

    Note:
        - the model must be given as a parameter because otherwise the tf.function wrapper will permanently
        associate the network variables with the graph, preventing us from clearing the memory

    Args:
        batch:                  the batch of data to learn on
        joint:                  the network (with both a policy and a value head
        distribution:           the distribution the network predicts
        continuous_control:     whether the distribution is continuous
        clip_values:            whether the value function should be clipped
        clipping_bound:         the bounds of the clipping
        gradient_clipping:      the clipping bound of the gradients, if None, no clipping is performed
        c_value:                the weight of the value functions loss
        c_entropy:              the weight of the entropy regularization
        is_recurrent:           whether the network is recurrent

    Returns:
        gradients, mean_entropym , mean_policy_loss, mean_value_loss
    """
    with tf.GradientTape() as tape:
        state_batch = {fname: f for fname, f in batch.items() if fname in Sensation.sense_names}
        old_values = batch["value"]

        policy_output, value_output = joint(state_batch, training=True)

        if continuous_control:
            # if action space is continuous, calculate PDF at chosen action value
            action_probabilities = distribution.log_probability(batch["action"], *policy_output)
        else:
            # if the action space is discrete, extract the probabilities of actions actually chosen
            action_probabilities = distribution.log_probability(batch["action"], policy_output)

        # calculate the three loss components
        policy_loss = loss.policy_loss(action_prob=action_probabilities,
                                       old_action_prob=batch["action_prob"],
                                       advantage=batch["advantage"],
                                       mask=batch["mask"],
                                       clipping_bound=clipping_bound,
                                       is_recurrent=is_recurrent)
        value_loss = loss.value_loss(value_predictions=tf.squeeze(value_output, axis=-1),
                                     old_values=old_values,
                                     returns=batch["return"],
                                     mask=batch["mask"],
                                     clip=clip_values,
                                     clipping_bound=clipping_bound,
                                     is_recurrent=is_recurrent)
        entropy = loss.entropy_bonus(policy_output=policy_output,
                                     distribution=distribution)

        # combine weighted losses
        total_loss = policy_loss + tf.multiply(c_value, value_loss) - tf.multiply(c_entropy, entropy)

    # calculate the gradient of the joint model based on total loss
    gradients = tape.gradient(total_loss, joint.trainable_variables)

    # clip gradients to avoid gradient explosion and stabilize learning
    if gradient_clipping is not None:
        gradients, _ = tf.clip_by_global_norm(gradients, gradient_clipping)

    entropy, policy_loss, value_loss = tf.reduce_mean(entropy), tf.reduce_mean(policy_loss), tf.reduce_mean(value_loss)

    return gradients, entropy, policy_loss, value_loss



import numpy as np
from numpy import ndarray as arr


def increment_mean_var(old_mean: arr, old_var: arr, new_mean: arr, new_var: arr, n: int, other_n: int = 1):
    """Incremental update of mean and variance.

    Variance based on parallel algorithm at https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance.
    """
    next_n = (n + other_n)
    mean = (n / next_n) * old_mean + (other_n / next_n) * new_mean

    delta = new_mean - old_mean
    m_a = old_var * (n - 1)
    m_b = new_var * (other_n - 1)
    m2 = m_a + m_b + delta ** 2 * n * other_n / (n + other_n)
    variance = m2 / (n + other_n - 1)

    return mean, variance


def ignore_none(func, sequence):
    """Apply function to a sequence ignoring all its Nones."""
    no_nones = [e for e in sequence if e is not None]
    if no_nones:
        return func(no_nones)

    return None


def mean_fill_nones(sequence, dtype=float):
    """In a sequence of numbers, replace Nones by the non-None boundary mean."""

    sequence = np.array(sequence, dtype=dtype)

    while np.any(np.isnan(sequence)):
        left = (np.isnan(sequence)).nonzero()[0] - 1
        right = (np.isnan(sequence)).nonzero()[0] + 1

        while np.any(np.isnan(sequence[left])):
            left[np.isnan(sequence[left])] -= 1

        while np.any(np.isnan(sequence[right])):
            right[np.isnan(sequence[right])] += 1

        sequence[np.isnan(sequence)] = (sequence[right] + sequence[left]) / 2

    return sequence


if __name__ == '__main__':
    means = [1, 2, 3, None, None, 6, 7, None, 9]
    print(mean_fill_nones(means))



import numpy as np

from angorapy.utilities.statistics import increment_mean_var


def test_incremental_mean_var():
    n_samples = 100000
    sample_dims = 10

    samples = np.array([np.random.randn(1, sample_dims) for _ in range(n_samples)])

    mean, var, n = samples[0], np.zeros((1, sample_dims,)), 1
    for s in samples[1:]:
        s_var = np.zeros((sample_dims,))
        mean, var = increment_mean_var(mean, var, s, s_var, n)
        n += 1

    np_mean = np.mean(samples, axis=0)
    np_var = np.var(samples, axis=0)

    assert np.allclose(mean, np_mean)
    assert np.allclose(var, np_var)



import itertools
import json
import os

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from angorapy.common.const import PATH_TO_EXPERIMENTS, QUALITATIVE_COLOR_PALETTE
from mpl_toolkits.axes_grid1.inset_locator import zoomed_inset_axes
from mpl_toolkits.axes_grid1.inset_locator import mark_inset


experiment_ids = [['1675028736765791', '1674983643322591'],
                  ['1674985163288377', '1674983177286330'],
                  ['1674343261520188', '1674308148646663', '1674074113967956', '1673350499432390']]  # compare distributions
names = ["beta", "gaussian", "multicategorical"]
# names = ["beta", "multicategorical"]
reward_developments = {}
reward_bands = {}

cosucc_developments = {}
cosucc_bands = {}

evaluation_rewards = {}
evaluation_reward_histograms = {}

for i, group in enumerate(experiment_ids):
    exp_name = names[i]
    reward_developments[exp_name] = []
    cosucc_developments[exp_name] = []
    evaluation_rewards[exp_name] = []
    evaluation_reward_histograms[exp_name] = []

    for id in group:
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "meta.json")) as f:
            meta = json.load(f)
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "progress.json")) as f:
            progress = json.load(f)
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "statistics.json")) as f:
            statistics = json.load(f)
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "evaluation.json")) as f:
            evaluation = json.load(f)

        reward_developments[exp_name].append(progress["rewards"]["mean"])
        cosucc_developments[exp_name].append(statistics["auxiliary_performances"]["consecutive_goals_reached"]["mean"])
        evaluation_rewards[exp_name] += evaluation["auxiliary_performances"]["consecutive_goals_reached"]
        evaluation_reward_histograms[exp_name].append(
            np.histogram(evaluation["auxiliary_performances"]["consecutive_goals_reached"], bins=50))

    evaluation_reward_histograms[exp_name] = np.mean([c for c, b in evaluation_reward_histograms[exp_name]], axis=0), \
    evaluation_reward_histograms[exp_name][0][1]
    reward_developments[exp_name] = np.array(
        [trace[:min(map(len, reward_developments[exp_name]))] for trace in reward_developments[exp_name]])
    cosucc_developments[exp_name] = np.array(
        [trace[:min(map(len, cosucc_developments[exp_name]))] for trace in cosucc_developments[exp_name]])

    rd_mean = np.mean(reward_developments[exp_name], axis=0)
    cd_mean = np.mean(cosucc_developments[exp_name], axis=0)
    rd_std = np.std(reward_developments[exp_name], axis=0)
    cd_std = np.std(cosucc_developments[exp_name], axis=0)
    reward_bands[exp_name] = (rd_mean - rd_std, rd_mean + rd_std)
    cosucc_bands[exp_name] = (cd_mean - cd_std, cd_mean + cd_std)

ax1 = plt.subplot2grid((1, 3), (0, 0), colspan=2)
ax2 = plt.subplot2grid((1, 3), (0, 2), colspan=1)
axins = zoomed_inset_axes(ax1, 2, loc=1) # zoom = 2

x_max = len(list(reward_developments.items())[0][1][0])
for i, (name, rewards) in enumerate(reward_developments.items()):
    x_max = max(x_max, len(rewards[0]))
    ax1.plot(np.mean(cosucc_developments[name], axis=0), label=name, color=QUALITATIVE_COLOR_PALETTE[i])
    ax1.fill_between(range(len(cosucc_bands[name][0])), cosucc_bands[name][0], cosucc_bands[name][1], alpha=0.2)

    axins.plot(np.mean(cosucc_developments[name], axis=0), label=name, color=QUALITATIVE_COLOR_PALETTE[i])
    axins.fill_between(range(len(cosucc_bands[name][0])), cosucc_bands[name][0], cosucc_bands[name][1], alpha=0.2)

    # ax2.hist(
    #     evaluation_reward_histograms[name][1][:-1],
    #     evaluation_reward_histograms[name][1],
    #     weights=evaluation_reward_histograms[name][0],
    #     edgecolor="white", linewidth=0.5, alpha=0.8
    # )
    # ax2.set_xlabel("Consecutive Goals Reached")
    # ax2.set_ylabel("Number of Episodes")
axins.set_xlim(0, 100.)
axins.set_ylim(-5., 10.)
plt.xticks(visible=False)
plt.yticks(visible=False)
mark_inset(ax1, axins, loc1=2, loc2=4, fc="none", ec="0.5")
# plt.draw()
df = pd.DataFrame(
    {"group": itertools.chain(*[[name] * len(dp) for name, dp in evaluation_rewards.items()]),
     "Consecutive Goals Reached": np.concatenate([dp for name, dp in evaluation_rewards.items()])}
)

sns.boxplot(data=df, x="group", y="Consecutive Goals Reached", medianprops={"color": "red"}, flierprops={"marker": "x"},
            fliersize=1, ax=ax2, palette={name: QUALITATIVE_COLOR_PALETTE[i] for i, name in enumerate(names)},
            showmeans=True, meanprops={"marker": "o", "markerfacecolor": "white", "markeredgecolor": "black"})
ax2.set_xlabel("")


ax1.set_xlabel("Cycle")
ax1.set_ylabel("Avg. Consecutive Goals Reached")
ax1.legend()

ax1.set_xlim(0, x_max)
ax1.set_ylim(-5)

plt.gcf().set_size_inches(12, 4)
plt.subplots_adjust(wspace=0.3, right=0.995, left=0.05, top=0.99)
plt.show()
# plt.savefig("../../../docs/figures/manipulate-asymmetry-ablation.pdf", format="pdf", bbox="tight")



import os
from typing import List

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import tensorflow as tf
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from analysis.visualization import Visualizer
from utilities.plotting import transparent_cmap

os.chdir("../../../")

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

tf.random.set_seed(69420)

# ANALYZERS
analyzer = Visualizer.from_saved_model("storage/pretrained/tanh_c.h5")

# FEATURE MAPS
fids = [1, 2, 3, 4, 5, 6, 7]
fmaxis = analyzer.preferred_stimulus("conv2d_2", feature_ids=fids)

axs: List[Axes]
fig, axs = plt.subplots(1, 7)
fig: Figure
fig.set_size_inches((15, 6))

i = 0
for ax in axs:
    ax.imshow(fmaxis[i], cmap="gray")
    ax.set_xticks([])
    ax.set_yticks([])

    i+=1

plt.savefig("docs/figures/fmaxis.pdf", format="pdf", bbox_inches='tight')
plt.show()



#!/usr/bin/env python
"""Convolutional components/networks."""
import keras_cortex.layers
import tensorflow as tf


def _residual_stack(x, filters, blocks, stride1=2, name=None):
    """A set of stacked residual blocks.
    Args:
      x: input tensor.
      filters: integer, filters of the bottleneck layer in a block.
      blocks: integer, blocks in the stacked blocks.
      stride1: default 2, stride of the first layer in the first block.
      name: string, stack label.
    Returns:
      Output tensor for the stacked blocks.
    """
    x = tfm.vision.layers.ResidualBlock(filters, strides=stride1, name=name + "_block1", use_projection=True)(x)
    for i in range(2, blocks + 1):
        x = tfm.vision.layers.ResidualBlock(
            filters, strides=1, use_projection=False, name=name + "_block" + str(i)
        )(x)
    return x


# VISUAL ENCODING


class OpenAIEncoder(tf.keras.Model):

    def __init__(self, shape, n_cameras=1, mode="pose", **kwargs):
        super().__init__(**kwargs)

        assert mode in ["pose", "classification"]
        self.mode = mode

        self.n_cameras = n_cameras

        self.rescale = tf.keras.layers.Rescaling(1.0 / 255)
        self.resnet_encoder = _build_openai_resnets(shape=shape,
                                                    batch_size=kwargs.get("batch_size"))

        self.dense = tf.keras.layers.Dense(128)
        self.relu = tf.keras.layers.Activation("relu")
        self.flatten = tf.keras.layers.Flatten()
        self.softmax = keras_cortex.layers.SpatialSoftmax()

        if mode == "pose":
            self.pos_dense = tf.keras.layers.Dense(3, kernel_initializer=tf.keras.initializers.GlorotUniform(),
                                                   bias_initializer=tf.keras.initializers.Constant(0),
                                                   kernel_regularizer=tf.keras.regularizers.L2(0.001),
                                                   bias_regularizer=tf.keras.regularizers.L2(0.001))
            self.rot_dense = tf.keras.layers.Dense(4, kernel_initializer=tf.keras.initializers.GlorotUniform(),
                                                   bias_initializer=tf.keras.initializers.Constant(0),
                                                   kernel_regularizer=tf.keras.regularizers.L2(0.001),
                                                   bias_regularizer=tf.keras.regularizers.L2(0.001))
        elif mode == "classification":
            self.data_augmentation = tf.keras.Sequential(
                [
                    tf.keras.layers.RandomFlip("horizontal"),
                    tf.keras.layers.RandomRotation(0.1),
                ]
            )

            self.classifier = tf.keras.layers.Dense(1000, activation="softmax")

    def call(self, inputs, training=None, mask=None):
        per_camera_output = []

        inputs = self.rescale(inputs)
        inputs = tf.split(inputs, num_or_size_splits=self.n_cameras, axis=-1)

        for camera in inputs:

            if self.mode == "classification":
                camera = self.data_augmentation(camera, training=training)

            x = self.resnet_encoder(camera)
            x = self.softmax(x)
            x = self.flatten(x)
            per_camera_output.append(x)

        # fully connected
        if len(per_camera_output) > 1:
            x = tf.keras.layers.concatenate(per_camera_output, axis=-1)
        else:
            x = per_camera_output[0]
        x = self.dense(x)
        x = self.relu(x)

        # output
        if self.mode == "pose":
            pos = self.pos_dense(x)
            rot = self.rot_dense(x)
            rot = tf.math.l2_normalize(rot, axis=-1)  # ensure quaternions are quaternions

            outputs = tf.keras.layers.concatenate([pos, rot], axis=-1)
        elif self.mode == "classification":
            outputs = self.classifier(x)
        else:
            raise NotImplementedError("Unknown mode: {}".format(self.mode))

        return outputs


def _build_openai_resnets(shape, batch_size=None):
    """Shallow AlexNet Version. Original number of channels are too large for normal GPU memory.

    """
    inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(shape))

    # first layer
    x = tf.keras.layers.Conv2D(32, 5, 1, padding="valid")(inputs)  # 196 x 196
    x = tf.keras.layers.Activation("relu")(x)

    # second layer
    x = tf.keras.layers.Conv2D(32, 3, 1, padding="valid")(x)  # 195 x 195
    x = tf.keras.layers.Activation("relu")(x)

    # pooling
    x = tf.keras.layers.MaxPool2D(3, 3)(x)  # 64 x 64

    # resnet
    x = _residual_stack(x, filters=16, blocks=1, stride1=1, name="res1")
    x = _residual_stack(x, filters=32, blocks=2, stride1=2, name="res2")
    x = _residual_stack(x, filters=64, blocks=2, stride1=2, name="res3")
    x = _residual_stack(x, filters=64, blocks=2, stride1=2, name="res4")

    return tf.keras.Model(inputs=inputs, outputs=x, name="resnetblocks")


def _build_openai_encoder(shape, name="visual_component", batch_size=None):
    """Shallow AlexNet Version. Original number of channels are too large for normal GPU memory.

    """
    inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(shape))

    # first layer
    x = tf.keras.layers.Conv2D(32, 5, 1, padding="valid")(inputs)  # 196 x 196
    x = tf.keras.layers.Activation("relu")(x)

    # second layer
    x = tf.keras.layers.Conv2D(32, 3, 1, padding="valid")(x)  # 195 x 195
    x = tf.keras.layers.Activation("relu")(x)

    # pooling
    x = tf.keras.layers.MaxPool2D(3, 3)(x)  # 64 x 64

    # resnet
    x = _residual_stack(x, filters=16, blocks=1, stride1=1, name="res1")
    x = _residual_stack(x, filters=32, blocks=2, stride1=2, name="res2")
    x = _residual_stack(x, filters=64, blocks=2, stride1=2, name="res3")
    x = _residual_stack(x, filters=64, blocks=2, stride1=2, name="res4")

    # fully connected
    # x = keras_cortex.layers.SpatialSoftmax()(x)
    x = tf.keras.layers.Flatten()(x)
    x = tf.keras.layers.Dense(128)(x)
    x = tf.keras.layers.Activation("relu")(x)

    # output
    pos = tf.keras.layers.Dense(3, kernel_initializer=tf.keras.initializers.GlorotUniform(),
                                bias_initializer=tf.keras.initializers.Constant(0),
                                kernel_regularizer=tf.keras.regularizers.L2(0.001),
                                bias_regularizer=tf.keras.regularizers.L2(0.001))(x)
    rot = tf.keras.layers.Dense(4, kernel_initializer=tf.keras.initializers.GlorotUniform(),
                                bias_initializer=tf.keras.initializers.Constant(0),
                                kernel_regularizer=tf.keras.regularizers.L2(0.001),
                                bias_regularizer=tf.keras.regularizers.L2(0.001))(x)
    rot = tf.math.l2_normalize(rot, axis=-1)  # ensure quaternions are quaternions

    outputs = tf.concat([pos, rot], axis=-1)

    return tf.keras.Model(inputs=inputs, outputs=outputs, name=name)


def _build_openai_small_encoder(shape, out_shape, name="visual_component", batch_size=None):
    """Shallow AlexNet Version. Original number of channels are too large for normal GPU memory.

    Args:
        out_shape:
    """
    inputs = tf.keras.Input(shape=shape, batch_size=batch_size, name=name + "_input")

    # first layer
    x = tf.keras.layers.Conv2D(32, 5, 1, padding="valid")(inputs)  # 96 x 96
    x = tf.keras.layers.Activation("relu")(x)

    # second layer
    x = tf.keras.layers.Conv2D(32, 3, 1, padding="valid")(x)  # 94 x 94
    x = tf.keras.layers.Activation("relu")(x)

    # pooling
    x = tf.keras.layers.MaxPool2D(3, 3)(x)  # 31 x 31

    # resnet
    x = _residual_block(x, 16, 3, 3, name="res1")
    x = _residual_block(x, 32, 3, 3, name="res2")
    x = _residual_block(x, 64, 3, 3, name="res3")

    # fully connected
    x = tf.keras.layers.Activation("softmax")(x)
    x = tf.keras.layers.Flatten()(x)
    x = tf.keras.layers.Dense(128, kernel_regularizer=tf.keras.regularizers.L2(l2=0.001))(x)
    x = tf.keras.layers.Activation("relu")(x)

    # output
    x = tf.keras.layers.Dense(out_shape)(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)


def _build_alexnet_encoder(shape, batch_size=None, name="visual_component"):
    """Shallow AlexNet Version. Original number of channels are too large for normal GPU memory."""
    inputs = tf.keras.Input(batch_shape=(batch_size,) + tuple(shape))

    # first layer
    x = tf.keras.layers.Conv2D(32, 11, 4)(inputs)
    x = tf.keras.layers.Activation("tanh")(x)
    x = tf.keras.layers.MaxPool2D(3, 2)(x)

    # second layer
    x = tf.keras.layers.Conv2D(32, 5, 1, padding="same")(x)
    x = tf.keras.layers.Activation("tanh")(x)
    x = tf.keras.layers.MaxPool2D(3, 2)(x)

    # third layer
    x = tf.keras.layers.Conv2D(64, 3, 1, padding="same")(x)
    x = tf.keras.layers.Activation("tanh")(x)

    # fourth layer
    x = tf.keras.layers.Conv2D(64, 3, 1, padding="same")(x)
    x = tf.keras.layers.Activation("tanh")(x)

    # fifth layer
    x = tf.keras.layers.Conv2D(32, 3, 1, padding="same")(x)
    x = tf.keras.layers.Activation("tanh")(x)

    x = tf.keras.layers.MaxPool2D(3, 2)(x)

    # fully connected layers
    x = tf.keras.layers.Flatten()(x)
    x = tf.keras.layers.Dense(512)(x)
    x = tf.keras.layers.Activation("tanh")(x)
    x = tf.keras.layers.Dense(512)(x)
    x = tf.keras.layers.Activation("tanh")(x)

    return tf.keras.Model(inputs=inputs, outputs=x, name=name)


# VISUAL DECODING

def _build_visual_decoder(input_dim):
    model = tf.keras.Sequential((
        tf.keras.layers.Dense(512, input_dim=input_dim, activation="relu"),
        tf.keras.layers.Dense(512, activation="relu"),
        tf.keras.layers.Dense(1152, activation="relu"),

        tf.keras.layers.Reshape((6, 6, 32)),
        tf.keras.layers.Conv2DTranspose(32, 3, 2),
        tf.keras.layers.Conv2DTranspose(64, 3, 1, activation="relu", padding="same"),
        tf.keras.layers.Conv2DTranspose(64, 3, 1, activation="relu", padding="same"),
        tf.keras.layers.Conv2DTranspose(32, 3, 2),
        tf.keras.layers.Conv2DTranspose(32, 5, 1, activation="relu", padding="same"),
        tf.keras.layers.Conv2DTranspose(32, 3, 2),
        tf.keras.layers.Conv2DTranspose(3, 11, 4, activation="sigmoid"),
    ))

    model.build()
    return model


if __name__ == "__main__":
    conv_comp = OpenAIEncoder((64, 64, 3), 1, mode="classification")
    print(conv_comp(tf.random.normal((1, 64, 64, 3))).shape)
    conv_comp.summary(expand_nested=True, )
    #
    # print("\n\n\n")
    # conv_comp = _build_openai_encoder((200, 200, 3), 7)
    # conv_comp.summary()



import sys

import tensorflow as tf


@tf.function
def policy_loss(action_prob: tf.Tensor,
                old_action_prob: tf.Tensor,
                advantage: tf.Tensor,
                mask: tf.Tensor,
                clipping_bound: tf.Tensor,
                is_recurrent: bool) -> tf.Tensor:
    """Clipped objective as given in the PPO paper. Original objective is to be maximized, but this is the negated
    objective to be minimized! In the recurrent version a mask is calculated based on 0 values in the
    old_action_prob tensor. The mask is then applied in the mean operation of the loss.

    Args:
      action_prob (tf.Tensor): the probability of the action for the state under the current policy
      old_action_prob (tf.Tensor): the probability of the action taken given by the old policy during the episode
      advantage (tf.Tensor): the advantage that taking the action gives over the estimated state value

    Returns:
      the value of the objective function

    """
    ratio = tf.exp(action_prob - old_action_prob)
    clipped = tf.maximum(
        tf.math.multiply(ratio, -advantage),
        tf.math.multiply(tf.clip_by_value(ratio, 1 - clipping_bound, 1 + clipping_bound), -advantage)
    )

    if is_recurrent:
        # build and apply a mask over the probabilities (recurrent)
        clipped_masked = tf.where(mask, clipped, 1)
        return tf.reduce_sum(clipped_masked) / tf.reduce_sum(tf.cast(mask, tf.float32))
    else:
        return tf.reduce_mean(clipped)


@tf.function
def value_loss(value_predictions: tf.Tensor,
               old_values: tf.Tensor,
               returns: tf.Tensor,
               mask: tf.Tensor,
               clip: bool,
               clipping_bound: tf.Tensor,
               is_recurrent: bool) -> tf.Tensor:
    """Loss of the critic network as squared error between the prediction and the sampled future return. In the
    recurrent case a mask is calculated based on 0 values in the old_action_prob tensor. This mask is then applied
    in the mean operation of the loss.

    Args:
      value_predictions (tf.Tensor): value prediction by the current critic network
      old_values (tf.Tensor): value prediction by the old critic network during gathering
      returns (tf.Tensor): discounted return estimation
      clip (object): (Default value = True) value loss can be clipped by same range as policy loss

    Returns:
      squared error between prediction and return
    """
    error = tf.square(value_predictions - returns)

    if clip:
        # clips value error to reduce variance
        clipped_values = old_values + tf.clip_by_value(value_predictions - old_values, -clipping_bound, clipping_bound)
        clipped_error = tf.square(clipped_values - returns)
        error = tf.maximum(clipped_error, error)

    if is_recurrent:
        # apply mask over the old values
        error_masked = tf.where(mask, error, 1)  # masking with tf.where because inf * 0 = nan...
        return (tf.reduce_sum(error_masked) / tf.reduce_sum(tf.cast(mask, tf.float32))) * 0.5
    else:
        return tf.reduce_mean(error) * 0.5


@tf.function
def entropy_bonus(policy_output: tf.Tensor, distribution) -> tf.Tensor:
    """Entropy of policy output acting as regularization by preventing dominance of one action. The higher the
    entropy, the less probability mass lies on a single action, which would hinder exploration. We hence reduce
    the loss by the (scaled by c_entropy) entropy to encourage a certain degree of exploration.

    Args:
      policy_output (tf.Tensor): a tensor containing (batches of) probabilities for actions in the case of discrete
        actions or (batches of) means and standard deviations for continuous control.

    Returns:
      entropy bonus
    """
    return tf.reduce_mean(distribution.entropy(policy_output))



import tensorflow as tf


# @tf.function
def lif(
    x, tau_rc=0.02, tau_ref=0.004, v_th=1., gamma=0.01, validity_threshold=-20., js_threshold=30.
):
    js = x / gamma
    j_valid = js > validity_threshold

    # splitting tf.where in two steps to avoid gradient nans
    safe_js_z = tf.where(
        js < js_threshold,
        js,
        tf.ones_like(js)
    )

    z = tf.where(
        js < js_threshold,
        tf.math.log1p(tf.math.exp(safe_js_z)),
        js,
    ) * gamma

    # splitting tf.where in two steps to avoid gradient nans
    safe_zs = tf.where(
        j_valid,
        z,
        tf.ones_like(z)
    )

    q = tf.where(
        j_valid,
        tf.math.log1p(1. / safe_zs),
        -js - tf.math.log(gamma)
    )

    response = v_th / (tau_ref + tau_rc * q)

    return tf.cast(response, x.dtype)


class LiF(tf.keras.layers.Layer):

    def __init__(self, tau_rc=0.02, tau_ref=0.004, v_th=1, gamma=0.01, **kwargs):
        super().__init__(**kwargs)
        self.v_th = tf.cast(v_th, dtype=tf.float32)
        self.tau_ref = tf.cast(tau_ref, dtype=tf.float32)
        self.tau_rc = tf.cast(tau_rc, dtype=tf.float32)
        self.gamma = tf.cast(gamma, dtype=tf.float32)

        self.js_threshold = tf.cast(30., dtype=tf.float32)
        self.validity_threshold = tf.cast(-20., dtype=tf.float32)

        self.supports_masking = True

    def call(self, inputs, **kwargs):
        return lif(
            inputs,
            gamma=self.gamma,
            v_th=self.v_th,
            tau_ref=self.tau_ref,
            tau_rc=self.tau_rc,
            js_threshold=self.js_threshold,
            validity_threshold=self.validity_threshold,
        )

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = {
            "v_th": self.v_th,
            "tau_ref": self.tau_ref,
            "tau_rc": self.tau_rc,
            "gamma": self.gamma
        }

        base_config = super().get_config()

        return dict(list(base_config.items()) + list(config.items()))


if __name__ == '__main__':
    from matplotlib import pyplot as plt

    inputs = tf.linspace(-2, 10, 100)

    for i in range(10):
        activation = LiF(tau_rc=0.02 / (i + 1))
        outputs = activation(tf.expand_dims(inputs, 0))

        plt.plot(inputs, tf.squeeze(outputs))

    plt.show()



from angorapy import make_task
from angorapy.common.policies import BetaPolicyDistribution, MultiCategoricalPolicyDistribution, \
    CategoricalPolicyDistribution
from angorapy.utilities.model_utils import is_recurrent_model
from angorapy.utilities.core import flatten

from angorapy.models import get_model_builder


def perform_test_on_model(model_name):
    cont_env = make_task("LunarLanderContinuous-v2")
    cont_distr = BetaPolicyDistribution(cont_env)

    discrete_env = make_task("LunarLander-v2")
    discrete_distr = CategoricalPolicyDistribution(discrete_env)

    multi_discrete_env = make_task("ManipulateBlockDiscrete-v0")
    multi_discrete_distr = MultiCategoricalPolicyDistribution(multi_discrete_env)

    build_model = get_model_builder(model=model_name, model_type="ffn", shared=False)
    build_shared_model = get_model_builder(model=model_name, model_type="ffn", shared=True)
    build_recurrent_model = get_model_builder(model=model_name, model_type="lstm", shared=False)
    build_shared_recurrent_model = get_model_builder(model=model_name, model_type="lstm", shared=True)

    for env, distr in [(cont_env, cont_distr), (discrete_env, discrete_distr), (multi_discrete_env, multi_discrete_distr)]:
        for model in [build_model, build_shared_model, build_recurrent_model, build_shared_recurrent_model]:
            _, _, joint = model(env, distr, bs=1, sequence_length=1)
            state = env.reset()[0]
            prepared_state = state.with_leading_dims(time=is_recurrent_model(joint)).dict_as_tf()
            policy_out = flatten(joint(prepared_state, training=False))

            predicted_distribution_parameters, value = policy_out[:-1], policy_out[-1]
            # from the action distribution sample an action and remember both the action and its probability
            action, action_probability = distr.act(*predicted_distribution_parameters)


def test_simple():
    """Test simple model."""
    perform_test_on_model("simple")


def test_wider():
    """Test simple model."""
    perform_test_on_model("wider")


def test_deeper():
    """Test deeper model."""
    perform_test_on_model("deeper")



#!/usr/bin/env python
"""Evaluate a loaded agent on a task."""
import argparse
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from angorapy.analysis.investigators import Investigator

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
from angorapy.tasks.envs.dexterity.core import BaseShadowHandEnv
from angorapy.tasks.envs.dexterity.consts import FINGERTIP_SITE_NAMES

import time

from angorapy.agent.ppo_agent import PPOAgent
from angorapy.common.const import BASE_SAVE_PATH, PATH_TO_EXPERIMENTS

from angorapy import make_task
import tensorflow as tf
import robotapper

import gymnasium as gym

tf.get_logger().setLevel('INFO')

parser = argparse.ArgumentParser(description="Inspect an episode of an agent.")
parser.add_argument("id", type=int, nargs="?", help="id of the agent, defaults to newest", default=None)
parser.add_argument("--env", type=str, nargs="?", help="force testing environment", default="")
parser.add_argument("--state", type=str, help="state, either iteration or 'best'", default="best")
parser.add_argument("--force-case-circulation", action="store_true", help="circle through goal definitions")
parser.add_argument("--freeze-wrist", action="store_true", help="prevent wrist movements")
parser.add_argument("--hide-targets", action="store_true", help="do not show visualization of targets")
parser.add_argument("--rcon", type=str, help="reward configuration", default=None)
parser.add_argument("--act-stochastic", action="store_true", help="keep stochasticity in decisions")

args = parser.parse_args()

scale_the_substeps = False

if args.state not in ["b", "best", "last"]:
    args.state = int(args.state)

if args.id is None:
    ids = map(int, [d for d in os.listdir(BASE_SAVE_PATH) if os.path.isdir(os.path.join(BASE_SAVE_PATH, d))])
    args.id = max(ids)

start = time.time()
agent = PPOAgent.from_agent_state(args.id, args.state, force_env_name=None if not args.env else args.env)
print(f"Agent {args.id} successfully loaded.")

try:
    tf.keras.utils.plot_model(agent.joint, to_file=f"{PATH_TO_EXPERIMENTS}/{args.id}/model.png", expand_nested=True,
                              show_shapes=True, dpi=300)
except:
    print("Could not create model plot.")

investigator = Investigator.from_agent(agent)
env = make_task(agent.env.spec.id, transformers=agent.env.transformers, render_mode="human")
if args.freeze_wrist:
    env.toggle_wrist_freezing()

substeps = "" if not hasattr(env.unwrapped, "sim") else f" with {env.unwrapped.sim.nsubsteps} substeps"
print(f"Evaluating on {env.unwrapped.spec.id}{substeps}.")
print(f"Environment has the following transformers: {env.transformers}")

# if args.env != "":
#     env = make_env(args.env, args.rcon)
# elif scale_the_substeps:
#     parts = env.env.unwrapped.spec.id.split("-")
#     new_name = parts[0] + "Fine" + "-" + parts[1]
#     print(new_name)
#     env = make_env(new_name, args.rcon)

if isinstance(env.unwrapped, BaseShadowHandEnv):
    env.set_delta_t_simulation(0.002)
    env.set_original_n_substeps_to_sspcs()
    env.change_color_scheme("default")

if not args.force_case_circulation or ("Reach" not in env.unwrapped.spec.id):
    for i in range(1000):
        investigator.render_episode(env, act_confidently=not args.act_stochastic)
# else:
#     env = make_env("ReachFFAbsolute-v0", transformers=agent.env.transformers)
#     if args.freeze_wrist:
#         env.env.toggle_wrist_freezing()
#
#     for i in range(1000):
#         print(f"\nNew Episode, finger {FINGERTIP_SITE_NAMES[i % 4]}")
#         env.forced_finger = i % 4
#         env.env.forced_finger = i % 4
#         env.unwrapped.forced_finger = i % 4
#         investigator.render_episode(env, slow_down=False)


from pathlib import Path
from typing import Dict, Tuple


JOINT_GROUP: Dict[str, Tuple[str, ...]] = {
    "wrist": ("WRJ1", "WRJ0"),
    "thumb": ("THJ4", "THJ3", "THJ2", "THJ1", "THJ0"),
    "first": ("FFJ3", "FFJ2", "FFJ1", "FFJ0"),
    "middle": ("MFJ3", "MFJ2", "MFJ1", "MFJ0"),
    "ring": ("RFJ3", "RFJ2", "RFJ1", "RFJ0"),
    "little": ("LFJ4", "LFJ3", "LFJ2", "LFJ1", "LFJ0"),
}

FINGERTIP_BODIES: Tuple[str, ...] = (
    # Important: the order of these names should not be changed.
    "thdistal",
    "ffdistal",
    "mfdistal",
    "rfdistal",
    "lfdistal",
)

FINGERTIP_COLORS: Tuple[Tuple[float, float, float], ...] = (
    # Important: the order of these colors should not be changed.
    (0.8, 0.2, 0.8),  # Purple.
    (0.8, 0.2, 0.2),  # Red.
    (0.2, 0.8, 0.8),  # Cyan.
    (0.2, 0.2, 0.8),  # Blue.
    (0.8, 0.8, 0.2),  # Yellow.
)

MODEL_XML = Path(__file__).resolve().parent / "bodies" / "gym_shadowhand.xml"



import itertools
import json
import os

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from angorapy.common.const import PATH_TO_EXPERIMENTS, QUALITATIVE_COLOR_PALETTE

experiment_ids = [['1674975602294059', '1674975602446141', '1671749718306373'],
                  ['1674343261520188', '1674308148646663', '1674074113967956', '1674074113731734', '1673350499432390']]
names = ["symmetric", "asymmetric"]

# experiment_ids = [['1674343261520188', '1674308148646663', '1674074113967956', '1674074113731734', '1673350499432390']]
# names = ["IHOM Agents"]

reward_developments = {}
reward_bands = {}

cosucc_developments = {}
cosucc_bands = {}

evaluation_rewards = {}
evaluation_reward_histograms = {}

for i, group in enumerate(experiment_ids):
    exp_name = names[i]
    reward_developments[exp_name] = []
    cosucc_developments[exp_name] = []
    cosucc_bands[exp_name] = []
    evaluation_rewards[exp_name] = []
    evaluation_reward_histograms[exp_name] = []

    for id in group:
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "meta.json")) as f:
            meta = json.load(f)
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "progress.json")) as f:
            progress = json.load(f)
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "statistics.json")) as f:
            statistics = json.load(f)

        try:
            with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "evaluation.json")) as f:
                evaluation = json.load(f)

            evaluation_rewards[exp_name] += evaluation["auxiliary_performances"]["consecutive_goals_reached"]
            evaluation_reward_histograms[exp_name].append(
                np.histogram(evaluation["auxiliary_performances"]["consecutive_goals_reached"], bins=50))
        except:
            pass

        window_size_one_sided = 5
        cosucc_traj = statistics["auxiliary_performances"]["consecutive_goals_reached"]["mean"]
        cosucc_developments[exp_name].append(np.array([
            np.mean(np.array(cosucc_traj)[max(i - window_size_one_sided, 0):min(i + window_size_one_sided, len(cosucc_traj))]) for i in
            range(len(cosucc_traj))
        ]))
        cosucc_bands[exp_name].append(np.array([
            np.std(np.array(cosucc_traj)[
                    max(i - window_size_one_sided, 0):min(i + window_size_one_sided, len(cosucc_traj))]) for i in
            range(len(cosucc_traj))
        ]))

    try:
        evaluation_reward_histograms[exp_name] = np.mean([c for c, b in evaluation_reward_histograms[exp_name]],
                                                         axis=0), evaluation_reward_histograms[exp_name][0][1]
    except:
        pass
    # reward_developments[exp_name] = np.array(
    #     [trace[:min(map(len, reward_developments[exp_name]))] for trace in reward_developments[exp_name]])
    # cosucc_developments[exp_name] = np.array(
    #     [trace[:min(map(len, cosucc_developments[exp_name]))] for trace in cosucc_developments[exp_name]])

ax1 = plt.subplot2grid((1, 3), (0, 0), colspan=3)
# ax2 = plt.subplot2grid((1, 3), (0, 2), colspan=1)

for i, (name, rewards) in enumerate(cosucc_developments.items()):
    for j in range(len(cosucc_developments[name])):
        ax1.plot(cosucc_developments[name][j], label=name if j == 0 else None, color=QUALITATIVE_COLOR_PALETTE[i])
        ax1.fill_between(range(len(cosucc_bands[name][j])),
                         cosucc_developments[name][j] - cosucc_bands[name][j],
                         cosucc_developments[name][j] + cosucc_bands[name][j], alpha=0.2, color=QUALITATIVE_COLOR_PALETTE[i])

    # ax2.hist(
    #     evaluation_reward_histograms[name][1][:-1],
    #     evaluation_reward_histograms[name][1],
    #     weights=evaluation_reward_histograms[name][0],
    #     edgecolor="white", linewidth=0.5, alpha=0.8
    # )
    # ax2.set_xlabel("Consecutive Goals Reached")
    # ax2.set_ylabel("Number of Episodes")

df = pd.DataFrame(
    {"group": itertools.chain(*[[name] * len(dp) for name, dp in evaluation_rewards.items()]),
     "Consecutive Goals Reached": np.concatenate([dp for name, dp in evaluation_rewards.items()])}
)

# sns.boxplot(data=df, x="group", y="Consecutive Goals Reached", medianprops={"color": "red"}, flierprops={"marker": "x"},
#             fliersize=1, ax=ax2, palette={name: QUALITATIVE_COLOR_PALETTE[i] for i, name in enumerate(names)},
#             showmeans=True, meanprops={"marker": "o", "markerfacecolor": "white", "markeredgecolor": "black"})
# ax2.set_xlabel("")

ax1.set_xlabel("Cycle")
ax1.set_ylabel("Avg. Consecutive Goals Reached")
ax1.legend(loc="upper left")

ax1.set_xlim(0)

plt.gcf().set_size_inches(12, 4)
plt.subplots_adjust(wspace=0.3, right=0.995, left=0.05, top=0.99)
plt.show()
# plt.savefig("../../../docs/figures/manipulate-asymmetry-ablation.pdf", format="pdf", bbox="tight")



import os

from angorapy.agent.utils import extract_discrete_action_probabilities

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import gymnasium as gym
import numpy as np
import tensorflow as tf
from scipy.stats import norm, entropy, beta

from angorapy.common.policies import GaussianPolicyDistribution, CategoricalPolicyDistribution, BetaPolicyDistribution
import pytest


# GAUSSIAN
def test_gaussian_pdf():
    distro = GaussianPolicyDistribution(gym.make("LunarLanderContinuous-v2"))

    x = tf.convert_to_tensor([[2, 3], [4, 3], [2, 1]], dtype=tf.float32)
    mu = tf.convert_to_tensor([[2, 1], [1, 3], [2, 2]], dtype=tf.float32)
    sig = tf.convert_to_tensor([[2, 2], [1, 2], [2, 1]], dtype=tf.float32)

    result_reference = np.prod(norm.pdf(x, loc=mu, scale=sig), axis=-1)
    result_pdf = distro.probability(x, mu, sig).numpy()
    result_log_pdf = np.exp(distro.log_probability(x, mu, np.log(sig)).numpy())

    assert np.allclose(result_reference, result_pdf), "Gaussian PDF returns wrong Result"
    assert np.allclose(result_pdf, result_log_pdf), "Gaussian Log PDF returns wrong Result"


def test_gaussian_entropy():
    distro = GaussianPolicyDistribution(gym.make("LunarLanderContinuous-v2"))

    mu = tf.convert_to_tensor([[2.0, 3.0], [2.0, 1.0]], dtype=tf.float32)
    sig = tf.convert_to_tensor([[1.0, 1.0], [1.0, 5.0]], dtype=tf.float32)

    result_reference = np.sum(norm.entropy(loc=mu, scale=sig), axis=-1)
    result_log = distro.entropy([mu, np.log(sig)]).numpy()
    result = distro._entropy_from_params(sig).numpy()

    assert np.allclose(result_reference, result), "Gaussian entropy returns wrong result"
    assert np.allclose(result_log, result_reference), "Gaussian entropy from log returns wrong result"


# BETA
def test_beta_pdf():
    distro = BetaPolicyDistribution(gym.make("LunarLanderContinuous-v2"))

    x = tf.convert_to_tensor([[0.2, 0.3], [0.4, 0.3], [0.2, 0.1]], dtype=tf.float32)
    alphas = tf.convert_to_tensor([[2, 1], [1, 3], [2, 2]], dtype=tf.float32)
    betas = tf.convert_to_tensor([[2, 2], [1, 2], [2, 1]], dtype=tf.float32)

    result_reference = np.prod(beta.pdf(distro._scale_sample_to_distribution_range(x), alphas, betas), axis=-1)
    result_pdf = distro.probability(x, alphas, betas).numpy()
    result_log_pdf = np.exp(distro.log_probability(x, alphas, betas).numpy())

    assert np.allclose(result_reference, result_log_pdf), "Beta Log PDF returns wrong Result"
    assert np.allclose(result_reference, result_pdf), "Beta PDF returns wrong Result"


def test_beta_entropy():
    distro = BetaPolicyDistribution(gym.make("LunarLanderContinuous-v2"))

    alphas = tf.convert_to_tensor([[2, 1], [1, 3], [2, 2]], dtype=tf.float32)
    betas = tf.convert_to_tensor([[2, 2], [1, 2], [2, 1]], dtype=tf.float32)

    result_reference = np.sum(beta.entropy(alphas, betas), axis=-1)
    result_pdf = distro._entropy_from_params((alphas, betas)).numpy()

    assert np.allclose(result_reference, result_pdf), "Beta PDF returns wrong Result"


# CATEGORICAL
def test_categorical_entropy():
    distro = CategoricalPolicyDistribution(gym.make("CartPole-v1"))

    probs = tf.convert_to_tensor([[0.1, 0.4, 0.2, 0.25, 0.05],
                                  [0.1, 0.4, 0.2, 0.2, 0.1],
                                  [0.1, 0.35, 0.3, 0.24, 0.01]], dtype=tf.float32)

    result_reference = [entropy(probs[i]) for i in range(len(probs))]
    result_log = distro._entropy_from_log_pmf(np.log(probs)).numpy()
    result = distro._entropy_from_pmf(probs).numpy()

    assert np.allclose(result_reference, result), "Discrete entropy returns wrong result"
    assert np.allclose(result_log, result_reference), "Discrete entropy from log returns wrong result"


def test_extract_discrete_action_probabilities():
    # no recurrence
    action_probs = tf.convert_to_tensor([[1, 5], [3, 7], [7, 2], [8, 4], [0, 2], [4, 5], [4, 2], [7, 5]])
    actions = tf.convert_to_tensor([1, 0, 1, 1, 0, 0, 0, 1])
    result_reference = tf.convert_to_tensor([5, 3, 2, 4, 0, 4, 4, 5])
    result = extract_discrete_action_probabilities(action_probs, actions)

    assert tf.reduce_all(tf.equal(result, result_reference)).numpy().item()


def test_extract_discrete_action_probabilities_with_recurrence():
    tf.config.experimental_run_functions_eagerly(True)

    # with recurrence
    action_probs = tf.convert_to_tensor(
        [[[1, 5], [1, 5]], [[3, 7], [3, 7]], [[7, 2], [7, 2]], [[8, 4], [8, 4]], [[0, 2], [0, 2]], [[4, 5], [4, 5]],
         [[4, 2], [4, 2]], [[7, 5], [7, 5]]])
    actions = tf.convert_to_tensor([[1, 1], [0, 0], [1, 1], [1, 1], [0, 0], [0, 0], [0, 0], [1, 1]])
    result_reference = tf.convert_to_tensor([[5, 5], [3, 3], [2, 2], [4, 4], [0, 0], [4, 4], [4, 4], [5, 5]])
    result = extract_discrete_action_probabilities(action_probs, actions)

    assert tf.reduce_all(tf.equal(result, result_reference)).numpy().item()